from marshmallow import fields
from marshmallow.schema import Schema
from marshmallow.validate import OneOf


class NgrokTunnelSchema(Schema):
    name = fields.String(
        metadata=dict(
            description='Tunnel friendly name or auto-generated name',
            example='tcp-8080-my-tunnel',
        )
    )

    protocol = fields.String(
        allow_none=False,
        attribute='proto',
        validate=OneOf(['tcp', 'udp', 'http']),
        metadata=dict(
            description='Tunnel protocol',
            example='tcp',
        ),
    )

    url = fields.String(
        attribute='public_url',
        required=True,
        metadata=dict(
            description='Public URL to the ngrok tunnel',
            example='tcp://8.tcp.ngrok.io:12345',
        )
    )



from platypush.backend.http.app.mixins import MessageType
from platypush.message.event import Event

from . import WSRoute, logger
from ..utils import send_message


class WSEventProxy(WSRoute):
    """
    Websocket event proxy mapped to ``/ws/events``.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, subscriptions=[self.events_channel], **kwargs)

    @classmethod
    def app_name(cls) -> str:
        return 'events'

    @classmethod
    @property
    def events_channel(cls) -> str:
        return cls.get_channel('events')

    @classmethod
    def publish(cls, data: MessageType, *_) -> None:
        super().publish(data, cls.events_channel)

    def on_message(self, message):
        try:
            event = Event.build(message)
            assert isinstance(event, Event), f'Expected {Event}, got {type(event)}'
        except Exception as e:
            logger.info('Could not build event from %s: %s', message, e)
            logger.exception(e)
            return

        send_message(event, wait_for_response=False)

    def run(self) -> None:
        for msg in self.listen():
            try:
                evt = Event.build(msg.data)
            except Exception as e:
                logger.warning('Error parsing event: %s: %s', msg, e)
                continue

            self.send(evt)



from abc import ABC, abstractmethod
from logging import getLogger
from typing import Iterable

from ._model import TorrentSearchResult


class TorrentSearchProvider(ABC):
    """
    Base class for torrent search providers.
    """

    def __init__(self, **kwargs):
        if 'enabled' in kwargs:
            self.enabled = bool(kwargs['enabled'])
        elif 'disabled' in kwargs:
            self.enabled = not bool(kwargs['disabled'])
        else:
            self.enabled = self.default_enabled()

        self.logger = getLogger(self.__class__.__name__)

    @abstractmethod
    def _search(self, query: str, *args, **kwargs) -> Iterable[TorrentSearchResult]:
        """
        Inner search method. This method should be implemented by subclasses.

        :param query: Query string, torrent name or partial name.
        """
        return []

    @classmethod
    @abstractmethod
    def provider_name(cls) -> str:
        """
        :return: Name of the provider, which can be used to identify it within
            the :class:`platypush.plugins.torrent.TorrentPlugin`.
        """

    def search(self, query: str, *args, **kwargs) -> Iterable[TorrentSearchResult]:
        """
        Perform a search of torrents.

        :param query: Query string, torrent name or partial name.
        """
        if not self.enabled:
            self.logger.debug(
                'Provider %s is disabled, skipping search', self.__class__.__name__
            )
            return []

        self.logger.debug("Searching for %r", query)
        return self._search(query, *args, **kwargs)

    @classmethod
    def default_enabled(cls) -> bool:
        """
        :return: True if the provider is enabled by default, False otherwise.
          Default: False.
        """
        return False


# vim:sw=4:ts=4:et:



import inspect
import logging
import os
from typing import Any, Callable, Mapping, Sequence, Tuple

from platypush.utils.manifest import Manifest

from ._types import StoppableThread

logger = logging.getLogger('platypush')


def _build_args(
    func: Callable, *args, **kwargs
) -> Tuple[Sequence[Any], Mapping[str, Any]]:
    spec = inspect.getfullargspec(func)
    func_args = args if spec.varargs else args[: len(spec.args)]
    func_kwargs = (
        kwargs
        if spec.varkw
        else {
            arg: kwargs[arg] for arg in [*spec.args, *spec.kwonlyargs] if arg in kwargs
        }
    )

    return func_args, func_kwargs


def exec_wrapper(f: Callable[..., Any], *args, **kwargs):
    """
    Utility function that runs a callable with its arguments, wraps its
    response into a ``Response`` object and handles errors/exceptions.
    """
    from platypush import Response

    func_args, func_kwargs = _build_args(f, *args, **kwargs)

    try:
        ret = f(*func_args, **func_kwargs)
        if isinstance(ret, Response):
            return ret

        return Response(output=ret)
    except Exception as e:
        logger.exception(e)
        return Response(errors=[str(e)])


# pylint: disable=too-few-public-methods
class ExtensionWithManifest:
    """
    This class models an extension with an associated manifest.json in the same
    folder.
    """

    def __init__(self, *_, **__):
        self._manifest = self.get_manifest()

    def get_manifest(self) -> Manifest:
        manifest_file = os.path.join(
            os.path.dirname(inspect.getfile(self.__class__)), 'manifest.json'
        )
        assert os.path.isfile(
            manifest_file
        ), f'The extension {self.__class__.__name__} has no associated manifest.json'

        return Manifest.from_file(manifest_file)


__all__ = [
    'ExtensionWithManifest',
    'StoppableThread',
    'exec_wrapper',
]



from abc import ABC, abstractmethod
from typing import List

from platypush.utils import get_plugin_name_by_class

from . import EntityManager


class WeatherEntityManager(EntityManager, ABC):
    """
    Base class for integrations that support weather reports.
    """

    def transform_entities(self, entities: List[dict], *, type: str):
        from platypush.entities.weather import Weather, WeatherForecast

        if not entities:
            return []

        plugin = get_plugin_name_by_class(self.__class__)

        if type == 'weather':
            # Current weather response
            weather = entities[0]
            weather.pop('time', None)
            return super().transform_entities(
                [
                    Weather(
                        id=f'{plugin}:weather',
                        name='Weather',
                        **weather,
                    )
                ]
            )

        # Weather forecast response
        if type == 'forecast':
            return super().transform_entities(
                [
                    WeatherForecast(
                        id=f'{plugin}:forecast',
                        name='Forecast',
                        forecast=entities,
                    )
                ]
            )

        raise AssertionError(f'Unexpected weather entity type: {type}')

    @abstractmethod
    def status(self, *_, **__):
        raise NotImplementedError



import asyncio
from dataclasses import dataclass
import threading
from typing import Optional

from bleak import BleakClient
from bleak.backends.device import BLEDevice


@dataclass
class BluetoothConnection:
    """
    A class to store information and context about a Bluetooth connection.
    """

    client: BleakClient
    device: BLEDevice
    loop: asyncio.AbstractEventLoop
    close_event: Optional[asyncio.Event] = None
    thread: Optional[threading.Thread] = None



import logging
import os
from typing import Any, Type

from PyOBEX.client import Client

from platypush.context import get_bus
from platypush.entities.bluetooth import BluetoothDevice
from platypush.message.event.bluetooth import (
    BluetoothConnectionFailedEvent,
    BluetoothDeviceEvent,
    BluetoothFileSentEvent,
    BluetoothFileTransferCancelledEvent,
    BluetoothFileTransferStartedEvent,
)

from platypush.plugins.bluetooth._legacy import LegacyManager
from platypush.plugins.bluetooth.model import ServiceClass


# pylint: disable=too-few-public-methods
class FileSender:
    """
    Wrapper for the Bluetooth file send OBEX service.
    """

    def __init__(self, scanner: LegacyManager):
        self._scanner = scanner
        self.logger = logging.getLogger(__name__)

    def send_file(
        self,
        file: str,
        device: str,
        data: bytes,
    ):
        """
        Send a file to a device.

        :param file: Name/path of the file to send.
        :param device: Name or address of the device to send the file to.
        :param data: File data.
        """

        dev = self._scanner.get_device(device)
        service = dev.known_services.get(ServiceClass.OBEX_OBJECT_PUSH)
        assert service, (
            f'The device {device} does not expose the service '
            f'{str(ServiceClass.OBEX_OBJECT_PUSH)}'
        )

        port = service.port
        client = self._connect(dev, port)
        self._post_event(BluetoothFileTransferStartedEvent, dev, file=file)
        self._send_file(client, dev, file, data)

    def _send_file(
        self,
        client: Client,
        dev: BluetoothDevice,
        file: str,
        data: bytes,
    ):
        filename = os.path.basename(file)

        try:
            client.put(filename, data)
            self._post_event(BluetoothFileSentEvent, dev, file=file)
        except Exception as e:
            self._post_event(
                BluetoothFileTransferCancelledEvent,
                dev,
                reason=str(e),
                file=file,
            )

            raise AssertionError(
                f'Failed to send file {file} to device {dev.address}: {e}'
            ) from e
        finally:
            client.disconnect()

    def _connect(self, dev: BluetoothDevice, port: str) -> Client:
        client = Client(dev.address, port)

        try:
            client.connect()
            assert (
                client.connection_id is not None
            ), 'Could not establish a connection to the device'
        except Exception as e:
            self._post_event(BluetoothConnectionFailedEvent, dev, reason=str(e))
            raise AssertionError(
                f'Connection to device {dev.address} failed: {e}'
            ) from e

        return client

    def _post_event(
        self,
        event_type: Type[BluetoothDeviceEvent],
        device: BluetoothDevice,
        **event_args: Any,
    ):
        get_bus().post(event_type.from_device(device, **event_args))



import datetime
import logging

from multiprocessing import Queue as PQueue
from queue import Empty, Queue as TQueue
from typing import Dict, List, Optional, Union
from uuid import UUID

from platypush.plugins import RunnablePlugin, action
from platypush.plugins.chat import ChatPlugin
from platypush.schemas.telegram import (
    TelegramChatSchema,
    TelegramFileSchema,
)
from platypush.utils import wait_for_either

from ._bridge import ResultBridge
from ._model import Command, Resource, END_OF_SERVICE
from ._service import TelegramService
from ._utils import dump_msg, dump_user


class TelegramPlugin(ChatPlugin, RunnablePlugin):
    """
    Plugin to programmatically send Telegram messages through a Telegram bot.
    In order to send messages to contacts, groups or channels you'll first need
    to register a bot. To do so:

        1. Open a Telegram conversation with the `@BotFather
           <https://telegram.me/BotFather>`_.
        2. Send ``/start`` followed by ``/newbot``. Choose a display name and a
           username for your bot.
        3. Copy the provided API token in the configuration of this plugin.
        4. Open a conversation with your newly created bot.

    """

    _DEFAULT_TIMEOUT = 20

    def __init__(
        self,
        api_token: str,
        authorized_chat_ids: Optional[List[Union[str, int]]] = None,
        **kwargs,
    ):
        """
        :param api_token: API token as returned by the `@BotFather
            <https://telegram.me/BotFather>`_
        :param authorized_chat_ids: Optional list of chat_id/user_id which are
            authorized to send messages to the bot. If nothing is specified
            then no restrictions are applied.
        """
        super().__init__(**kwargs)
        self._authorized_chat_ids = set(authorized_chat_ids or [])
        self._api_token = api_token
        self._cmd_queue = PQueue()
        self._result_queue = PQueue()
        self._service: Optional[TelegramService] = None
        self._response_queues: Dict[UUID, TQueue] = {}
        self._result_bridge: Optional[ResultBridge] = None

        # Set httpx logging to WARNING, or it will log a line for every request
        logging.getLogger("httpx").setLevel(logging.WARNING)

    def _exec(
        self, cmd: str, *args, timeout: Optional[float] = _DEFAULT_TIMEOUT, **kwargs
    ):
        assert self._service, "Telegram service not running"
        cmd_obj = Command(cmd, args=args, kwargs=kwargs, timeout=timeout)
        self._response_queues[cmd_obj.id] = TQueue()
        self._cmd_queue.put_nowait(cmd_obj)

        try:
            result = self._response_queues[cmd_obj.id].get(timeout=timeout)
        except (TimeoutError, Empty) as e:
            raise TimeoutError(f"Timeout while executing command {cmd}") from e
        finally:
            self._response_queues.pop(cmd_obj.id, None)

        assert not isinstance(
            result, Exception
        ), f'Error while executing command {cmd}: {result}'

        return result

    @property
    def response_queues(self):
        return self._response_queues

    @property
    def result_queue(self):
        return self._result_queue

    @action
    def send_message(  # pylint: disable=arguments-differ
        self,
        chat_id: Union[str, int],
        text: str,
        *_,
        parse_mode: Optional[str] = None,
        disable_web_page_preview: bool = False,
        disable_notification: bool = False,
        reply_to_message_id: Optional[int] = None,
        timeout: Optional[float] = _DEFAULT_TIMEOUT,
        **__,
    ) -> dict:
        """
        Send a message to a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``. In order to get your own
            Telegram chat_id open a conversation with `@IDBot
            <https://telegram.me/IDBot>`_ and type ``/start`` followed by
            ``/getid``. Similar procedures also exist to get a group or channel
            chat_id - just Google for "Telegram get channel/group chat_id".
        :param text: Text to be sent.
        :param parse_mode: Set to 'Markdown' or 'HTML' to send either Markdown
            or HTML content.
        :param disable_web_page_preview: If True then web previews for URLs
            will be disabled.
        :param disable_notification: If True then no notification will be sent
            to the users.
        :param reply_to_message_id: If set then the message will be sent as a
            response to the specified message.
        :param timeout: Request timeout (default: 20 seconds).
        :return: .. schema:: telegram.TelegramMessageSchema
        """
        return dump_msg(
            self._exec(
                'send_message',
                chat_id=chat_id,
                text=text,
                parse_mode=parse_mode,
                disable_web_page_preview=disable_web_page_preview,
                disable_notification=disable_notification,
                reply_to_message_id=reply_to_message_id,
                timeout=timeout,
            )
        )

    @action
    def send_photo(
        self,
        chat_id: Union[str, int],
        file_id: Optional[int] = None,
        url: Optional[str] = None,
        path: Optional[str] = None,
        caption: Optional[str] = None,
        parse_mode: Optional[str] = None,
        disable_notification: bool = False,
        reply_to_message_id: Optional[int] = None,
        timeout: Optional[float] = _DEFAULT_TIMEOUT,
    ) -> dict:
        """
        Send a picture to a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``. In order to get your own
            Telegram chat_id open a conversation with `@IDBot
            <https://telegram.me/IDBot>`_ and type ``/start`` followed by
            ``/getid``. Similar procedures also exist to get a group or channel
            chat_id - just Google for "Telegram get channel/group chat_id".
        :param file_id: Set it if the file already exists on Telegram servers
            and has a file_id. Note that you'll have to specify either
            ``file_id``, ``url`` or ``path``.
        :param url: Set it if you want to send a file from a remote URL. Note
            that you'll have to specify either ``file_id``, ``url`` or
            ``path``.
        :param path: Set it if you want to send a file from the local
            filesystem. Note that you'll have to specify either ``file_id``,
            ``url`` or ``path``.
        :param caption: Optional caption for the picture.
        :param parse_mode: Set to 'Markdown' or 'HTML' to send either Markdown
            or HTML content.
        :param disable_notification: If True then no notification will be sent
            to the users.
        :param reply_to_message_id: If set then the message will be sent as a
            response to the specified message.
        :param timeout: Upload timeout (default: 20 seconds)
        :return: .. schema:: telegram.TelegramMessageSchema
        """
        return dump_msg(
            self._exec(
                'send_photo',
                chat_id=chat_id,
                resource=Resource(file_id=file_id, url=url, path=path),
                resource_attr='photo',
                caption=caption,
                disable_notification=disable_notification,
                reply_to_message_id=reply_to_message_id,
                timeout=timeout,
                parse_mode=parse_mode,
            )
        )

    @action
    def send_audio(
        self,
        chat_id: Union[str, int],
        file_id: Optional[int] = None,
        url: Optional[str] = None,
        path: Optional[str] = None,
        caption: Optional[str] = None,
        performer: Optional[str] = None,
        title: Optional[str] = None,
        duration: Optional[float] = None,
        parse_mode: Optional[str] = None,
        disable_notification: bool = False,
        reply_to_message_id: Optional[int] = None,
        timeout: Optional[float] = _DEFAULT_TIMEOUT,
    ) -> dict:
        """
        Send audio to a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``. In order to get your own
            Telegram chat_id open a conversation with `@IDBot
            <https://telegram.me/IDBot>`_ and type ``/start`` followed by
            ``/getid``. Similar procedures also exist to get a group or channel
            chat_id - just Google for "Telegram get channel/group chat_id".
        :param file_id: Set it if the file already exists on Telegram servers
            and has a file_id. Note that you'll have to specify either
            ``file_id``, ``url`` or ``path``.
        :param url: Set it if you want to send a file from a remote URL. Note
            that you'll have to specify either ``file_id``, ``url`` or
            ``path``.
        :param path: Set it if you want to send a file from the local
            filesystem. Note that you'll have to specify either ``file_id``,
            ``url`` or ``path``.
        :param caption: Optional caption for the picture.
        :param performer: Optional audio performer.
        :param title: Optional audio title.
        :param duration: Duration of the audio in seconds.
        :param parse_mode: Set to 'Markdown' or 'HTML' to send either Markdown
            or HTML content.
        :param disable_notification: If True then no notification will be sent
            to the users.
        :param reply_to_message_id: If set then the message will be sent as a
            response to the specified message.
        :param timeout: Upload timeout (default: 20 seconds)
        :return: .. schema:: telegram.TelegramMessageSchema
        """
        return dump_msg(
            self._exec(
                'send_audio',
                chat_id=chat_id,
                resource=Resource(file_id=file_id, url=url, path=path),
                resource_attr='audio',
                caption=caption,
                disable_notification=disable_notification,
                performer=performer,
                title=title,
                duration=duration,
                reply_to_message_id=reply_to_message_id,
                timeout=timeout,
                parse_mode=parse_mode,
            )
        )

    @action
    def send_document(
        self,
        chat_id: Union[str, int],
        file_id: Optional[int] = None,
        url: Optional[str] = None,
        path: Optional[str] = None,
        filename: Optional[str] = None,
        caption: Optional[str] = None,
        parse_mode: Optional[str] = None,
        disable_notification: bool = False,
        reply_to_message_id: Optional[int] = None,
        timeout: Optional[float] = _DEFAULT_TIMEOUT,
    ) -> dict:
        """
        Send a document to a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``. In order to get your own
            Telegram chat_id open a conversation with `@IDBot
            <https://telegram.me/IDBot>`_ and type ``/start`` followed by
            ``/getid``. Similar procedures also exist to get a group or channel
            chat_id - just Google for "Telegram get channel/group chat_id".
        :param file_id: Set it if the file already exists on Telegram servers
            and has a file_id. Note that you'll have to specify either
            ``file_id``, ``url`` or ``path``.
        :param url: Set it if you want to send a file from a remote URL. Note
            that you'll have to specify either ``file_id``, ``url`` or
            ``path``.
        :param path: Set it if you want to send a file from the local
            filesystem. Note that you'll have to specify either ``file_id``,
            ``url`` or ``path``.
        :param filename: Name of the file as it will be shown in Telegram.
        :param caption: Optional caption for the picture.
        :param parse_mode: Set to 'Markdown' or 'HTML' to send either Markdown
            or HTML content.
        :param disable_notification: If True then no notification will be sent
            to the users.
        :param reply_to_message_id: If set then the message will be sent as a
            response to the specified message.
        :param timeout: Upload timeout (default: 20 seconds)
        :return: .. schema:: telegram.TelegramMessageSchema
        """
        return dump_msg(
            self._exec(
                'send_document',
                chat_id=chat_id,
                resource=Resource(file_id=file_id, url=url, path=path),
                resource_attr='document',
                filename=filename,
                caption=caption,
                disable_notification=disable_notification,
                reply_to_message_id=reply_to_message_id,
                timeout=timeout,
                parse_mode=parse_mode,
            )
        )

    @action
    def send_video(
        self,
        chat_id: Union[str, int],
        file_id: Optional[int] = None,
        url: Optional[str] = None,
        path: Optional[str] = None,
        duration: Optional[int] = None,
        caption: Optional[str] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
        parse_mode: Optional[str] = None,
        disable_notification: bool = False,
        reply_to_message_id: Optional[int] = None,
        timeout: Optional[float] = _DEFAULT_TIMEOUT,
    ) -> dict:
        """
        Send a video to a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``. In order to get your own
            Telegram chat_id open a conversation with `@IDBot
            <https://telegram.me/IDBot>`_ and type ``/start`` followed by
            ``/getid``. Similar procedures also exist to get a group or channel
            chat_id - just Google for "Telegram get channel/group chat_id".
        :param file_id: Set it if the file already exists on Telegram servers
            and has a file_id. Note that you'll have to specify either
            ``file_id``, ``url`` or ``path``.
        :param url: Set it if you want to send a file from a remote URL. Note
            that you'll have to specify either ``file_id``, ``url`` or
            ``path``.
        :param path: Set it if you want to send a file from the local
            filesystem. Note that you'll have to specify either ``file_id``,
            ``url`` or ``path``.
        :param duration: Duration in seconds.
        :param caption: Optional caption for the picture.
        :param width: Video width.
        :param height: Video height.
        :param parse_mode: Set to 'Markdown' or 'HTML' to send either Markdown
            or HTML content.
        :param disable_notification: If True then no notification will be sent
            to the users.
        :param reply_to_message_id: If set then the message will be sent as a
            response to the specified message.
        :param timeout: Upload timeout (default: 20 seconds)
        :return: .. schema:: telegram.TelegramMessageSchema
        """
        return dump_msg(
            self._exec(
                'send_video',
                chat_id=chat_id,
                resource=Resource(file_id=file_id, url=url, path=path),
                resource_attr='video',
                duration=duration,
                caption=caption,
                width=width,
                height=height,
                disable_notification=disable_notification,
                reply_to_message_id=reply_to_message_id,
                timeout=timeout,
                parse_mode=parse_mode,
            )
        )

    @action
    def send_animation(
        self,
        chat_id: Union[str, int],
        file_id: Optional[int] = None,
        url: Optional[str] = None,
        path: Optional[str] = None,
        duration: Optional[int] = None,
        caption: Optional[str] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
        parse_mode: Optional[str] = None,
        disable_notification: bool = False,
        reply_to_message_id: Optional[int] = None,
        timeout: Optional[float] = _DEFAULT_TIMEOUT,
    ) -> dict:
        """
        Send an animation (GIF or H.264/MPEG-4 AVC video without sound) to a
        chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``. In order to get your own
            Telegram chat_id open a conversation with `@IDBot
            <https://telegram.me/IDBot>`_ and type ``/start`` followed by
            ``/getid``. Similar procedures also exist to get a group or channel
            chat_id - just Google for "Telegram get channel/group chat_id".
        :param file_id: Set it if the file already exists on Telegram servers
            and has a file_id. Note that you'll have to specify either
            ``file_id``, ``url`` or ``path``.
        :param url: Set it if you want to send a file from a remote URL. Note
            that you'll have to specify either ``file_id``, ``url`` or
            ``path``.
        :param path: Set it if you want to send a file from the local
            filesystem. Note that you'll have to specify either ``file_id``,
            ``url`` or ``path``.
        :param duration: Duration in seconds.
        :param caption: Optional caption for the picture.
        :param width: Video width.
        :param height: Video height.
        :param parse_mode: Set to 'Markdown' or 'HTML' to send either Markdown
            or HTML content.
        :param disable_notification: If True then no notification will be sent
            to the users.
        :param reply_to_message_id: If set then the message will be sent as a
            response to the specified message.
        :param timeout: Upload timeout (default: 20 seconds)
        :return: .. schema:: telegram.TelegramMessageSchema
        """
        return dump_msg(
            self._exec(
                'send_animation',
                chat_id=chat_id,
                resource=Resource(file_id=file_id, url=url, path=path),
                resource_attr='animation',
                duration=duration,
                caption=caption,
                width=width,
                height=height,
                disable_notification=disable_notification,
                reply_to_message_id=reply_to_message_id,
                timeout=timeout,
                parse_mode=parse_mode,
            )
        )

    @action
    def send_voice(
        self,
        chat_id: Union[str, int],
        file_id: Optional[int] = None,
        url: Optional[str] = None,
        path: Optional[str] = None,
        caption: Optional[str] = None,
        duration: Optional[float] = None,
        parse_mode: Optional[str] = None,
        disable_notification: bool = False,
        reply_to_message_id: Optional[int] = None,
        timeout: Optional[float] = _DEFAULT_TIMEOUT,
    ) -> dict:
        """
        Send audio to a chat as a voice file. For this to work, your audio must
        be in an .ogg file encoded with OPUS (other formats may be sent as
        Audio or Document).

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``. In order to get your own
            Telegram chat_id open a conversation with `@IDBot
            <https://telegram.me/IDBot>`_ and type ``/start`` followed by
            ``/getid``. Similar procedures also exist to get a group or channel
            chat_id - just Google for "Telegram get channel/group chat_id".
        :param file_id: Set it if the file already exists on Telegram servers
            and has a file_id. Note that you'll have to specify either
            ``file_id``, ``url`` or ``path``.
        :param url: Set it if you want to send a file from a remote URL. Note
            that you'll have to specify either ``file_id``, ``url`` or ``path``.
        :param path: Set it if you want to send a file from the local
            filesystem. Note that you'll have to specify either ``file_id``,
            ``url`` or ``path``.
        :param caption: Optional caption for the picture.
        :param duration: Duration of the voice in seconds.
        :param parse_mode: Set to 'Markdown' or 'HTML' to send either Markdown
            or HTML content.
        :param disable_notification: If True then no notification will be sent
            to the users.
        :param reply_to_message_id: If set then the message will be sent as a
            response to the specified message.
        :param timeout: Upload timeout (default: 20 seconds)
        :return: .. schema:: telegram.TelegramMessageSchema
        """
        return dump_msg(
            self._exec(
                'send_voice',
                chat_id=chat_id,
                resource=Resource(file_id=file_id, url=url, path=path),
                resource_attr='voice',
                caption=caption,
                disable_notification=disable_notification,
                duration=duration,
                reply_to_message_id=reply_to_message_id,
                timeout=timeout,
                parse_mode=parse_mode,
            )
        )

    @action
    def send_video_note(
        self,
        chat_id: Union[str, int],
        file_id: Optional[int] = None,
        url: Optional[str] = None,
        path: Optional[str] = None,
        duration: Optional[int] = None,
        disable_notification: bool = False,
        reply_to_message_id: Optional[int] = None,
        timeout: Optional[float] = _DEFAULT_TIMEOUT,
    ) -> dict:
        """
        Send a video note to a chat. As of v.4.0, Telegram clients support
        rounded square mp4 videos of up to 1 minute long.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``. In order to get your own
            Telegram chat_id open a conversation with `@IDBot
            <https://telegram.me/IDBot>`_ and type ``/start`` followed by
            ``/getid``. Similar procedures also exist to get a group or channel
            chat_id - just Google for "Telegram get channel/group chat_id".
        :param file_id: Set it if the file already exists on Telegram servers
            and has a file_id. Note that you'll have to specify either
            ``file_id``, ``url`` or ``path``.
        :param url: Set it if you want to send a file from a remote URL. Note
            that you'll have to specify either ``file_id``, ``url`` or
            ``path``.
        :param path: Set it if you want to send a file from the local
            filesystem. Note that you'll have to specify either ``file_id``,
            ``url`` or ``path``.
        :param duration: Duration in seconds.
        :param disable_notification: If True then no notification will be sent
            to the users.
        :param reply_to_message_id: If set then the message will be sent as a
            response to the specified message.
        :param timeout: Upload timeout (default: 20 seconds)
        :return: .. schema:: telegram.TelegramMessageSchema
        """
        return dump_msg(
            self._exec(
                'send_video_note',
                chat_id=chat_id,
                resource=Resource(file_id=file_id, url=url, path=path),
                resource_attr='video',
                duration=duration,
                disable_notification=disable_notification,
                reply_to_message_id=reply_to_message_id,
                timeout=timeout,
            )
        )

    @action
    def send_location(
        self,
        chat_id: Union[str, int],
        latitude: float,
        longitude: float,
        disable_notification: bool = False,
        reply_to_message_id: Optional[int] = None,
        timeout: Optional[float] = _DEFAULT_TIMEOUT,
    ) -> dict:
        """
        Send a location to a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``. In order to get your own
            Telegram chat_id open a conversation with `@IDBot
            <https://telegram.me/IDBot>`_ and type ``/start`` followed by
            ``/getid``. Similar procedures also exist to get a group or channel
            chat_id - just Google for "Telegram get channel/group chat_id".
        :param latitude: Latitude
        :param longitude: Longitude
        :param disable_notification: If True then no notification will be sent
            to the users.
        :param reply_to_message_id: If set then the message will be sent as a
            response to the specified message.
        :param timeout: Upload timeout (default: 20 seconds)
        :return: .. schema:: telegram.TelegramMessageSchema
        """
        return dump_msg(
            self._exec(
                'send_location',
                chat_id=chat_id,
                latitude=latitude,
                longitude=longitude,
                disable_notification=disable_notification,
                reply_to_message_id=reply_to_message_id,
                timeout=timeout,
            )
        )

    @action
    def send_venue(
        self,
        chat_id: Union[str, int],
        latitude: float,
        longitude: float,
        title: str,
        address: str,
        foursquare_id: Optional[str] = None,
        foursquare_type: Optional[str] = None,
        disable_notification: bool = False,
        reply_to_message_id: Optional[int] = None,
        timeout: Optional[float] = _DEFAULT_TIMEOUT,
    ) -> dict:
        """
        Send the address of a venue to a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``. In order to get your own
            Telegram chat_id open a conversation with `@IDBot
            <https://telegram.me/IDBot>`_ and type ``/start`` followed by
            ``/getid``. Similar procedures also exist to get a group or channel
            chat_id - just Google for "Telegram get channel/group chat_id".
        :param latitude: Latitude
        :param longitude: Longitude
        :param title: Venue name.
        :param address: Venue address.
        :param foursquare_id: Foursquare ID.
        :param foursquare_type: Foursquare type.
        :param disable_notification: If True then no notification will be sent
            to the users.
        :param reply_to_message_id: If set then the message will be sent as a
            response to the specified message.
        :param timeout: Upload timeout (default: 20 seconds)
        :return: .. schema:: telegram.TelegramMessageSchema
        """
        return dump_msg(
            self._exec(
                'send_venue',
                chat_id=chat_id,
                latitude=latitude,
                longitude=longitude,
                title=title,
                address=address,
                foursquare_id=foursquare_id,
                foursquare_type=foursquare_type,
                disable_notification=disable_notification,
                reply_to_message_id=reply_to_message_id,
                timeout=timeout,
            )
        )

    @action
    def send_contact(
        self,
        chat_id: Union[str, int],
        phone_number: str,
        first_name: str,
        last_name: Optional[str] = None,
        vcard: Optional[str] = None,
        disable_notification: bool = False,
        reply_to_message_id: Optional[int] = None,
        timeout: Optional[float] = _DEFAULT_TIMEOUT,
    ) -> dict:
        """
        Send a contact to a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``. In order to get your own
            Telegram chat_id open a conversation with `@IDBot
            <https://telegram.me/IDBot>`_ and type ``/start`` followed by
            ``/getid``. Similar procedures also exist to get a group or channel
            chat_id - just Google for "Telegram get channel/group chat_id".
        :param phone_number: Phone number.
        :param first_name: First name.
        :param last_name: Last name.
        :param vcard: Additional contact info in vCard format (0-2048 bytes).
        :param disable_notification: If True then no notification will be sent
            to the users.
        :param reply_to_message_id: If set then the message will be sent as a
            response to the specified message.
        :param timeout: Upload timeout (default: 20 seconds)
        :return: .. schema:: telegram.TelegramMessageSchema
        """
        return dump_msg(
            self._exec(
                'send_contact',
                chat_id=chat_id,
                phone_number=phone_number,
                first_name=first_name,
                last_name=last_name,
                vcard=vcard,
                disable_notification=disable_notification,
                reply_to_message_id=reply_to_message_id,
                timeout=timeout,
            )
        )

    @action
    def get_file(self, file_id: str, timeout: Optional[float] = None) -> dict:
        """
        Get the info and URL of an uploaded file by file_id.

        :param file_id: File ID.
        :param timeout: Upload timeout (default: 20 seconds).
        :return: .. schema:: telegram.TelegramFileSchema
        """
        return dict(
            TelegramFileSchema().dump(self._exec('get_file', file_id, timeout=timeout))
        )

    @action
    def get_chat(
        self, chat_id: Union[int, str], timeout: Optional[float] = _DEFAULT_TIMEOUT
    ) -> dict:
        """
        Get the info about a Telegram chat.

        :param chat_id: Chat ID.
        :param timeout: Upload timeout (default: 20 seconds).
        """
        return dict(
            TelegramChatSchema().dump(self._exec('get_chat', chat_id, timeout=timeout))
        )

    @action
    def get_chat_user(
        self,
        chat_id: Union[int, str],
        user_id: int,
        timeout: Optional[float] = _DEFAULT_TIMEOUT,
    ) -> dict:
        """
        Get the info about a user connected to a chat.

        :param chat_id: Chat ID.
        :param user_id: User ID.
        :param timeout: Upload timeout (default: 20 seconds).
        """
        return dump_user(
            self._exec('get_chat_member', chat_id, user_id, timeout=timeout)
        )

    @action
    def get_chat_administrators(
        self, chat_id: Union[int, str], timeout: Optional[float] = _DEFAULT_TIMEOUT
    ) -> List[dict]:
        """
        Get the list of the administrators of a chat.

        :param chat_id: Chat ID.
        :param timeout: Upload timeout (default: 20 seconds).
        :return: .. schema:: telegram.TelegramUserSchema(many=True)
        """
        return [
            dump_user(user.user)
            for user in self._exec('get_chat_administrators', chat_id, timeout=timeout)
        ]

    @action
    def get_chat_members_count(
        self, chat_id: Union[int, str], timeout: Optional[float] = _DEFAULT_TIMEOUT
    ) -> int:
        """
        Get the number of users in a chat.

        :param chat_id: Chat ID.
        :param timeout: Upload timeout (default: 20 seconds).
        """
        return self._exec('get_chat_members_count', chat_id, timeout=timeout)

    @action
    def kick_chat_member(
        self,
        chat_id: Union[str, int],
        user_id: int,
        until_date: Optional[datetime.datetime] = None,
        timeout: Optional[float] = _DEFAULT_TIMEOUT,
    ):
        """
        Kick a user from a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``. In order to get your own
            Telegram chat_id open a conversation with `@IDBot
            <https://telegram.me/IDBot>`_ and type ``/start`` followed by
            ``/getid``. Similar procedures also exist to get a group or channel
            chat_id - just Google for "Telegram get channel/group chat_id".
        :param user_id: Unique user ID.
        :param until_date: End date for the ban.
        :param timeout: Request timeout (default: 20 seconds)
        """
        self._exec(
            'kick_chat_member',
            chat_id=chat_id,
            user_id=user_id,
            until_date=until_date,
            timeout=timeout,
        )

    @action
    def unban_chat_member(
        self,
        chat_id: Union[str, int],
        user_id: int,
        timeout: Optional[float] = _DEFAULT_TIMEOUT,
    ):
        """
        Lift the ban from a chat member.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``. In order to get your own
            Telegram chat_id open a conversation with `@IDBot
            <https://telegram.me/IDBot>`_ and type ``/start`` followed by
            ``/getid``. Similar procedures also exist to get a group or channel
            chat_id - just Google for "Telegram get channel/group chat_id".
        :param user_id: Unique user ID.
        :param timeout: Request timeout (default: 20 seconds)
        """
        self._exec(
            'unban_chat_member', chat_id=chat_id, user_id=user_id, timeout=timeout
        )

    @action
    def promote_chat_member(
        self,
        chat_id: Union[str, int],
        user_id: int,
        can_change_info: Optional[bool] = None,
        can_post_messages: Optional[bool] = None,
        can_edit_messages: Optional[bool] = None,
        can_delete_messages: Optional[bool] = None,
        can_invite_users: Optional[bool] = None,
        can_restrict_members: Optional[bool] = None,
        can_promote_members: Optional[bool] = None,
        can_pin_messages: Optional[bool] = None,
        timeout: Optional[float] = _DEFAULT_TIMEOUT,
    ):
        """
        Promote or demote a member.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``. In order to get your own
            Telegram chat_id open a conversation with `@IDBot
            <https://telegram.me/IDBot>`_ and type ``/start`` followed by
            ``/getid``. Similar procedures also exist to get a group or channel
            chat_id - just Google for "Telegram get channel/group chat_id".
        :param user_id: Unique user ID.
        :param can_change_info: Pass True if the user can change channel info.
        :param can_post_messages: Pass True if the user can post messages.
        :param can_edit_messages: Pass True if the user can edit messages.
        :param can_delete_messages: Pass True if the user can delete messages.
        :param can_invite_users: Pass True if the user can invite other users
            to the channel/group.
        :param can_restrict_members: Pass True if the user can restrict the
            permissions of other users.
        :param can_promote_members: Pass True if the user can promote mebmers.
        :param can_pin_messages: Pass True if the user can pin messages.
        :param timeout: Request timeout (default: 20 seconds)
        """
        self._exec(
            'promote_chat_member',
            chat_id=chat_id,
            user_id=user_id,
            can_change_info=can_change_info,
            can_post_messages=can_post_messages,
            can_edit_messages=can_edit_messages,
            can_delete_messages=can_delete_messages,
            can_invite_users=can_invite_users,
            can_restrict_members=can_restrict_members,
            can_promote_members=can_promote_members,
            can_pin_messages=can_pin_messages,
            timeout=timeout,
        )

    @action
    def set_chat_title(
        self,
        chat_id: Union[str, int],
        title: str,
        timeout: Optional[float] = _DEFAULT_TIMEOUT,
    ):
        """
        Set the title of a channel/group.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``. In order to get your own
            Telegram chat_id open a conversation with `@IDBot
            <https://telegram.me/IDBot>`_ and type ``/start`` followed by
            ``/getid``. Similar procedures also exist to get a group or channel
            chat_id - just Google for "Telegram get channel/group chat_id".
        :param title: New chat title.
        :param timeout: Request timeout (default: 20 seconds)
        """
        self._exec(
            'set_chat_title', chat_id=chat_id, description=title, timeout=timeout
        )

    @action
    def set_chat_description(
        self,
        chat_id: Union[str, int],
        description: str,
        timeout: Optional[float] = _DEFAULT_TIMEOUT,
    ):
        """
        Set the description of a channel/group.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``. In order to get your own
            Telegram chat_id open a conversation with `@IDBot
            <https://telegram.me/IDBot>`_ and type ``/start`` followed by
            ``/getid``. Similar procedures also exist to get a group or channel
            chat_id - just Google for "Telegram get channel/group chat_id".
        :param description: New chat description.
        :param timeout: Request timeout (default: 20 seconds)
        """
        self._exec(
            'set_chat_description',
            chat_id=chat_id,
            description=description,
            timeout=timeout,
        )

    @action
    def set_chat_photo(
        self,
        chat_id: Union[str, int],
        path: str,
        timeout: Optional[float] = _DEFAULT_TIMEOUT,
    ):
        """
        Set the photo of a channel/group.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``. In order to get your own
            Telegram chat_id open a conversation with `@IDBot
            <https://telegram.me/IDBot>`_ and type ``/start`` followed by
            ``/getid``. Similar procedures also exist to get a group or channel
            chat_id - just Google for "Telegram get channel/group chat_id".
        :param path: Path of the new image.
        :param timeout: Request timeout (default: 20 seconds)
        """
        self._exec(
            'set_chat_photo',
            chat_id=chat_id,
            resource=Resource(path=path),
            resource_attr='photo',
            timeout=timeout,
        )

    @action
    def delete_chat_photo(
        self, chat_id: Union[str, int], timeout: Optional[float] = _DEFAULT_TIMEOUT
    ):
        """
        Delete the photo of a channel/group.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``. In order to get your own
            Telegram chat_id open a conversation with `@IDBot
            <https://telegram.me/IDBot>`_ and type ``/start`` followed by
            ``/getid``. Similar procedures also exist to get a group or channel
            chat_id - just Google for "Telegram get channel/group chat_id".
        :param timeout: Request timeout (default: 20 seconds)
        """
        self._exec('delete_chat_photo', chat_id=chat_id, timeout=timeout)

    @action
    def pin_chat_message(
        self,
        chat_id: Union[str, int],
        message_id: int,
        disable_notification: Optional[bool] = None,
        timeout: Optional[float] = _DEFAULT_TIMEOUT,
    ):
        """
        Pin a message in a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``. In order to get your own
            Telegram chat_id open a conversation with `@IDBot
            <https://telegram.me/IDBot>`_ and type ``/start`` followed by
            ``/getid``. Similar procedures also exist to get a group or channel
            chat_id - just Google for "Telegram get channel/group chat_id".
        :param message_id: Message ID.
        :param disable_notification: If True then no notification will be sent
            to the users.
        :param timeout: Request timeout (default: 20 seconds)
        """
        self._exec(
            'pin_chat_message',
            chat_id=chat_id,
            message_id=message_id,
            disable_notification=disable_notification,
            timeout=timeout,
        )

    @action
    def unpin_chat_message(
        self, chat_id: Union[str, int], timeout: Optional[float] = _DEFAULT_TIMEOUT
    ):
        """
        Unpin the message of a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``. In order to get your own
            Telegram chat_id open a conversation with `@IDBot
            <https://telegram.me/IDBot>`_ and type ``/start`` followed by
            ``/getid``. Similar procedures also exist to get a group or channel
            chat_id - just Google for "Telegram get channel/group chat_id".
        :param timeout: Request timeout (default: 20 seconds)
        """
        self._exec('unpin_chat_message', chat_id=chat_id, timeout=timeout)

    @action
    def leave_chat(
        self, chat_id: Union[str, int], timeout: Optional[float] = _DEFAULT_TIMEOUT
    ):
        """
        Leave a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique
            identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed
            by ``/getid``. Similar procedures also exist to get a group or
            channel chat_id - just Google for "Telegram get channel/group
            chat_id".
        :param timeout: Request timeout (default: 20 seconds)
        """
        self._exec('leave_chat', chat_id=chat_id, timeout=timeout)

    def main(self):
        self._result_bridge = ResultBridge(self)
        self._result_bridge.start()

        while not self.should_stop():
            self._service = TelegramService(
                self._api_token,
                cmd_queue=self._cmd_queue,
                result_queue=self._result_queue,
                authorized_chat_ids=self._authorized_chat_ids,
            )

            self._service.start()
            wait_for_either(self._should_stop, self._service.stop_event)
            self.wait_stop(10)

    def stop(self):
        # Send an END_OF_SERVICE to the result queue to signal that the service
        # is stopping
        self._cmd_queue.put_nowait(END_OF_SERVICE)

        if self._service:
            self._service.stop()
            self._service = None

        if self._result_bridge and self._result_bridge.is_alive():
            self._result_bridge.join()
            self._result_bridge = None

        super().stop()


# vim:sw=4:ts=4:et:



import datetime
import os
import re
import requests
import threading
import xmlrpc.client

from pathlib import Path
from typing import List, Optional

from platypush.context import get_bus
from platypush.plugins import action
from platypush.plugins.torrent import TorrentPlugin
from platypush.message.event.torrent import (
    TorrentDownloadStartEvent,
    TorrentDownloadedMetadataEvent,
    TorrentDownloadProgressEvent,
    TorrentDownloadCompletedEvent,
    TorrentPausedEvent,
    TorrentResumedEvent,
    TorrentQueuedEvent,
    TorrentRemovedEvent,
    TorrentEvent,
)


class RtorrentPlugin(TorrentPlugin):
    """
    Plugin to interact search, download and manage torrents through RTorrent.

    You may prefer the built-in :class:`platypush.plugins.torrent.TorrentPlugin` over this one, unless you have heavy
    dependencies on RTorrent, as quite some extra configuration is required to enable RTorrent's RPC API -
    which is required to communicate with this integration.

    Configuration:

        - Install ``rtorrent`` on your system - on Debian/Ubuntu/Raspbian::

            apt-get install rtorrent

        - Configure the ``rtorrent`` XML/RPC interface, usually by adding the following lines to your
          ``~/.rtorrent.rc``:

            .. code-block:: yaml

              # Enable XML/RPC
              scgi_local = /home/user/.rpc.socket

        - Use a web server to bridge the RPC interface exposed by RTorrent over HTTP. Some configuration examples are
          available `here <https://github.com/rakshasa/rtorrent/wiki/RPC-Setup-XMLRPC>`_. I usually use ``lighttpd``
          because it's easy to configure and it comes with a built-in SCGI module. Install the server e.g. using
          ``apt``::

            apt-get install lighttpd

        - Create a base configuration file like this under e.g. ``~/.config/rtorrent/lighttpd.conf``:

          .. code-block:: python

            ### Base configuration
            server.modules = (
                "mod_indexfile",
                "mod_access",
                "mod_alias",
                "mod_redirect",
            )

            # Make sure that all the directories exist.

            # server.document-root isn't really needed, but lighttpd
            # won't start if it doesn't find a document root.
            server.document-root        = "/home/user/.local/share/rtorrent/html"
            server.upload-dirs          = ( "/home/user/.cache/uploads" )
            server.errorlog             = "/home/user/.local/log/rtorrent/error.log"
            server.pid-file             = "/home/user/.local/run/lighttpd.pid"
            server.username             = "your-user"
            server.groupname            = "your-group"
            server.port                 = 5000

            index-file.names            = ( "index.html" )

            ### Configure the RTorrent XML/RPC endpoint
            server.modules += ( "mod_scgi" )
            scgi.server = (
                # Bind an endpoint called /RPC2 to your local interface
                "/RPC2" =>
                  ( "127.0.0.1" =>
                    (
                      # Read from the RTorrent XML/RPC socket
                      "socket" => "/home/user/.rpc.socket",
                      "check-local" => "disable",
                      "disable-time" => 0,  # don't disable scgi if connection fails
                    )
                  )
              )

        - Start the HTTP service, and optionally enable it as a system/user service::

            lighttpd -f ~/.config/rtorrent/lighttpd.conf

        - Start RTorrent and check that the XML/RPC interface works:

          .. code-block:: bash

            $ xmlrpc localhost:8000 system.listMethods
            # Should return a list with all the methods exposed by RTorrent.
            $ xmlrpc localhost:5000 download_list
            Result:
            Array of 0 items:

        - It is advised to let the RTorrent instance run in e.g. ``screen`` or ``tmux`` on the server machine - it is
          more reliable than letting the plugin start/stop the instance, and you have an easy CLI interface to attach
          to manage/monitor your torrents.

        - In this example, the URL to configure in the plugin would be ``http://localhost:5000/RPC2``.

    """

    def __init__(
        self,
        url: str,
        poll_seconds: float = 5.0,
        download_dir: str = '~/.rtorrent/watch',
        **kwargs
    ):
        """
        :param url: HTTP URL that exposes the XML/RPC interface of RTorrent (e.g. ``http://localhost:5000/RPC2``).
        :param poll_seconds: How often the plugin will monitor for changes in the torrent state (default: 5 seconds).
        :param download_dir: Directory where torrents and metadata files will be downloaded
            (default: ``~/.rtorrent/watch``).
        """
        super().__init__(**kwargs)
        self.torrent_files_dir = os.path.abspath(os.path.expanduser(download_dir))
        Path(self.torrent_files_dir).mkdir(parents=True, exist_ok=True, mode=0o755)

        self._monitor_stop = threading.Event()
        self._monitor_thread: Optional[threading.Thread] = None
        self._last_status = {}
        self._torrent_urls = {}
        self._status_lock = threading.RLock()

        self.poll_seconds = poll_seconds
        self.url = url
        self.client = xmlrpc.client.Server(self.url)
        self.methods = set(self._list_methods())
        self.start_monitor()

    def _get_client(self) -> xmlrpc.client.Server:
        return xmlrpc.client.Server(self.url)

    def _fire_event(self, event: TorrentEvent, *_, **__):
        bus = get_bus()
        bus.post(event)

    def _process_events(self, status: dict, last_status: dict):
        if not status:
            self._fire_event(TorrentRemovedEvent(**last_status))
            return

        if not last_status:
            self._fire_event(TorrentQueuedEvent(**status))

        progress = status.get('progress', 0)
        name = status.get('name')
        start_date = status.get('start_date')
        finish_date = status.get('finish_date')
        is_active = status.get('is_active')

        if name and not last_status.get('name'):
            self._fire_event(TorrentDownloadedMetadataEvent(**status))

        if start_date and not last_status.get('start_date'):
            self._fire_event(TorrentDownloadStartEvent(**status))

        if is_active and not last_status.get('is_active'):
            self._fire_event(TorrentResumedEvent(**status))
        elif not is_active and last_status.get('is_active'):
            self._fire_event(TorrentPausedEvent(**status))

        if progress > 0 and progress > last_status.get('progress', 0):
            self._fire_event(TorrentDownloadProgressEvent(**status))

        if finish_date and not last_status.get('finish_date'):
            self._fire_event(TorrentDownloadCompletedEvent(**status))

    def _torrent_monitor(self, *_, **__):
        def thread():
            self.logger.info('Starting torrent monitoring')

            while not self._monitor_stop.is_set():
                try:
                    # noinspection PyUnresolvedReferences
                    statuses = self.status().output
                    last_statuses = self._last_status.copy()
                    self._last_status = statuses
                    torrent_hashes = set(statuses.keys()).union(last_statuses.keys())

                    for torrent_hash in torrent_hashes:
                        self._process_events(
                            statuses.get(torrent_hash, {}),
                            last_statuses.get(torrent_hash, {}),
                        )
                except Exception as e:
                    self.logger.warning('Error while monitoring torrent status')
                    self.logger.exception(e)
                finally:
                    self._monitor_stop.wait(timeout=self.poll_seconds)

            self.logger.info('Stopped torrent monitoring')

        return thread

    def _multicall(self, *args) -> List[list]:
        if 'd.multicall2' in self.methods:
            return self.client.d.multicall2('', *args)
        if 'd.multicall' in self.methods:
            return self.client.d.multicall(*args)

        raise AssertionError('No multicall method available on the rtorrent interface')

    @action
    def start_monitor(self):
        """
        Start monitoring the status of the RTorrent instance.
        """
        if self._monitor_thread and self._monitor_thread.is_alive():
            self.logger.info('Torrent monitoring already running')
            return

        self._monitor_stop.clear()
        self._monitor_thread = threading.Thread(target=self._torrent_monitor())
        self._monitor_thread.start()

    @action
    def stop_monitor(self):
        """
        Stop monitoring the status of the RTorrent instance.
        """
        if not (self._monitor_thread and self._monitor_thread.is_alive()):
            self.logger.info('Torrent monitoring already stopped')
        else:
            self._monitor_stop.set()
            self._monitor_thread.join(timeout=60.0)

        self._monitor_thread = None

    @action
    def download_torrent_file(self, torrent: str) -> str:
        """
        Download a torrent link to ``torrent_files_dir``.

        :param torrent: Torrent URL, magnet link or local file.
        :return: Path to the locally downloaded .torrent file.
        """
        if torrent.startswith('magnet:?'):
            # Magnet link: extract and download
            m = re.search(r'xt=urn:btih:([^&/]+)', torrent)
            assert m, 'Invalid magnet link: {}'.format(torrent)
            torrent_hash = m.group(1)
            torrent_file = os.path.join(
                self.torrent_files_dir, '{}.torrent'.format(torrent_hash)
            )

            with open(torrent_file, 'w') as f:
                f.write(
                    'd10:magnet-uri{length}:{info}e'.format(
                        length=len(torrent), info=torrent
                    )
                )

            self._torrent_urls[torrent_hash] = torrent
            return torrent_file

        if torrent.startswith('http://') or torrent.startswith('https://'):
            # HTTP resource
            info = requests.get(torrent).text
            torrent_file = os.path.join(self.torrent_files_dir, torrent.split('/')[-1])
            if not torrent_file.endswith('.torrent'):
                torrent_file += '.torrent'

            with open(torrent_file, 'w') as f:
                f.write(info)

            self._torrent_urls[torrent_file.split('.')[0]] = torrent
            return torrent_file

        # Local torrent file
        torrent_file = os.path.abspath(os.path.expanduser(torrent))
        assert os.path.isfile(torrent_file), 'No such torrent file: {}'.format(torrent)

        self._torrent_urls[os.path.basename(torrent_file).split('.')[0]] = (
            'file://' + torrent
        )
        return torrent_file

    @action
    def download(self, torrent: str, is_media: bool = False, *_, **__):
        """
        Download a torrent.

        :param torrent: Torrent to download. Supported formats:

            * Magnet URLs
            * Torrent URLs
            * Local torrent files

        :param is_media: Set it to true if you're downloading a media file that you'd like to stream as soon as the
            first chunks are available. If so, then the events and the status method will only include media files
        :return: The status of the torrent.
        """
        # noinspection PyUnresolvedReferences
        torrent_file = self.download_torrent_file(torrent).output
        client = self._get_client()
        client.load.start('', torrent_file)

    def _list_methods(self) -> List[str]:
        return self.client.system.listMethods()

    @action
    def list_methods(self) -> List[str]:
        """
        :return: The list of methods exposed by the RTorrent instance
        """
        return list(self.methods)

    @action
    def status(self, torrent: str = None) -> dict:
        """
        Get the status of the current transfers.

        :param torrent: Torrent hash.
        :returns: A dictionary:

          .. code-block:: json

            {
              "HASH1234567890": {
                "hash": "HASH1234567890",
                "name": "Your torrent name",
                "save_path": "/home/user/Downloads/Your torrent name",
                "is_active": true,
                "is_open": true,
                "completed_bytes": 666894336,
                "download_rate": 451345,
                "is_multi_file": true,
                "remaining_bytes": 1482827011,
                "size_bytes": 2149721347,
                "load_date": "2020-09-02T18:42:19",
                "peers": 0,
                "state": "paused",
                "start_date": "2020-09-02T18:42:19",
                "finish_date": null,
                "upload_rate": 143967,
                "progress": 31.0,
                "files": ["list", "of", "downloaded", "files"]
              }
            }

        """
        attrs = [
            'hash',
            'name',
            'save_path',
            'is_active',
            'is_open',
            'completed_bytes',
            'download_rate',
            'is_multi_file',
            'remaining_bytes',
            'size_bytes',
            'load_date',
            'peers',
            'start_date',
            'finish_date',
            'upload_rate',
        ]
        cmds = [
            'd.hash=',
            'd.name=',
            'd.directory=',
            'd.is_active=',
            'd.is_open=',
            'd.completed_bytes=',
            'd.down.rate=',
            'd.is_multi_file=',
            'd.left_bytes=',
            'd.size_bytes=',
            'd.load_date=',
            'd.peers_connected=',
            'd.timestamp.started=',
            'd.timestamp.finished=',
            'd.up.rate=',
        ]

        mappers = {
            'is_active': lambda v: bool(v),
            'is_open': lambda v: bool(v),
            'is_multi_file': lambda v: bool(v),
            'load_date': lambda v: datetime.datetime.fromtimestamp(v) if v else None,
            'start_date': lambda v: datetime.datetime.fromtimestamp(v) if v else None,
            'finish_date': lambda v: datetime.datetime.fromtimestamp(v) if v else None,
        }

        with self._status_lock:
            torrents = {
                info[0]: {
                    attr: mappers[attr](info[i]) if attr in mappers else info[i]
                    for i, attr in enumerate(attrs)
                }
                for info in self._multicall('', *cmds)
            }

        for torrent_id, info in torrents.items():
            torrents[torrent_id]['progress'] = round(
                100.0 * (info['completed_bytes'] / info['size_bytes']), 1
            )
            torrents[torrent_id]['url'] = self._torrent_urls.get(torrent_id, torrent_id)
            torrents[torrent_id]['is_paused'] = not info['is_active']
            torrents[torrent_id]['paused'] = not info[
                'is_active'
            ]  # Back compatibility with TorrentPlugin
            torrents[torrent_id]['size'] = info[
                'size_bytes'
            ]  # Back compatibility with TorrentPlugin
            torrents[torrent_id]['files'] = []

            if not info['is_open']:
                torrents[torrent_id]['state'] = 'stopped'
            elif not info['is_active']:
                torrents[torrent_id]['state'] = 'paused'
            else:
                torrents[torrent_id]['state'] = 'downloading'

            if info.get('save_path'):
                torrents[torrent_id]['files'] = (
                    [str(f) for f in Path(info['save_path']).rglob('*')]
                    if info.get('is_multi_file')
                    else info['save_path']
                )

        return torrents.get(torrent, {}) if torrent else torrents

    @action
    def open(self, torrent: str) -> dict:
        """
        Open a loaded torrent transfer.

        :param torrent: Torrent hash.
        :return: The status of the torrent.
        """
        self.client.d.open(torrent)
        return self.status(torrent).output

    @action
    def pause(self, torrent: str) -> dict:
        """
        Pause a torrent transfer.

        :param torrent: Torrent hash.
        :return: The status of the torrent.
        """
        self.client.d.pause(torrent)
        return self.status(torrent).output

    @action
    def resume(self, torrent) -> dict:
        """
        Resume a torrent transfer.

        :param torrent: Torrent hash.
        :return: The status of the torrent.
        """
        self.client.d.resume(torrent)
        return self.status(torrent).output

    @action
    def stop(self, torrent) -> dict:
        """
        Stop a torrent transfer.

        :param torrent: Torrent hash.
        :return: The status of the torrent.
        """
        self.client.d.stop(torrent)
        return self.status(torrent).output

    @action
    def remove(self, torrent):
        """
        Stop and remove a torrent transfer (without removing the downloaded files).

        :param torrent: Torrent hash.
        """
        self.client.d.stop(torrent)
        self.client.d.erase(torrent)

    @action
    def quit(self):
        """
        Terminate all the active transfers and quit the monitor.
        """
        # noinspection PyUnresolvedReferences
        torrents = list(self.status().output.keys()).copy()
        for torrent in torrents:
            self.remove(torrent)

        self.stop_monitor()

    @action
    def execute(self, method: str, *args, **kwargs):
        """
        Execute a raw command over the RTorrent RPC interface.

        :param method: Method name.
        :param args: Method arguments.
        :param kwargs: Method keyword-arguments.
        :return: Anything returned by the RPC method.
        """
        method = getattr(self.client, method)
        return method(*args, **kwargs)


# vim:sw=4:ts=4:et:



from enum import Enum
from typing import Iterable
from uuid import UUID

from platypush.entities import Entity
from platypush.entities.bluetooth import BluetoothDevice
from platypush.entities.switches import EnumSwitch
from platypush.plugins.bluetooth.model import ServiceClass
from platypush.plugins.bluetooth._plugins import BaseBluetoothPlugin


def _make_uuid(prefix: str) -> UUID:
    """
    Utility method to create a Switchbot characteristic UUID given a hex
    prefix.
    """
    return UUID(f'cba20{prefix}-224d-11e6-9fb8-0002a5d5c51b')


class Command(Enum):
    """
    Supported commands.
    """

    PRESS = b'\x57\x01\x00'
    ON = b'\x57\x01\x01'
    OFF = b'\x57\x01\x02'


class Characteristic(Enum):
    """
    GATT characteristic UUIDs supported by Switchbot devices.
    """

    TX = _make_uuid('002')
    RX = _make_uuid('003')
    SRV = _make_uuid('d00')


# pylint: disable=too-few-public-methods
class SwitchbotPlugin(BaseBluetoothPlugin):
    """
    Implements support for Switchbot devices.
    """

    def supports_device(self, device: BluetoothDevice) -> bool:
        return any(
            srv.service_class == ServiceClass.SWITCHBOT for srv in device.services
        )

    def _extract_entities(self, device: BluetoothDevice) -> Iterable[Entity]:
        return [
            EnumSwitch(
                id=f'{device.address}::switchbot',
                name='Switchbot',
                values=[cmd.name.lower() for cmd in Command],
                is_write_only=True,
            )
        ]

    def set(self, device: BluetoothDevice, value: str, **_) -> None:
        value = value.upper()
        cmd = getattr(Command, value, None)
        assert cmd, f'No such command: {value}. Available commands: {list(Command)}.'
        self._manager.write(device.address, cmd.value, Characteristic.TX.value)



from marshmallow import pre_load

from .._base import SystemBaseSchema


class DiskBaseSchema(SystemBaseSchema):
    """
    Base schema for disk stats.
    """

    @pre_load
    def pre_load(self, data: dict, **_) -> dict:
        data = super().pre_load(data)

        # Convert read/write/busy times from milliseconds to seconds
        for attr in ['read_time', 'write_time', 'busy_time']:
            if data.get(attr) is not None:
                data[attr] /= 1000

        # Normalize the percentage between 0 and 1
        if data.get('percent') is not None:
            data['percent'] /= 100

        return data



from enum import Enum
from socket import AddressFamily

from marshmallow import pre_load

from .._base import SystemBaseSchema


class NetworkInterfaceBaseSchema(SystemBaseSchema):
    """
    Base schema for network interface stats.
    """

    @pre_load
    def pre_load(self, data: dict, **_) -> dict:
        data = super().pre_load(data)

        # Custom attribute mappings
        for in_attr, out_attr in {
            'errin': 'errors_in',
            'errout': 'errors_out',
            'dropin': 'drop_in',
            'dropout': 'drop_out',
            'isup': 'is_up',
        }.items():
            if in_attr in data:
                data[out_attr] = data.pop(in_attr)

        # Serialize enum values
        for i, addr in enumerate(data.get('addresses', [])):
            if hasattr(addr, '_asdict'):
                addr = addr._asdict()
            if isinstance(addr.get('family'), AddressFamily):
                addr['family'] = addr['family'].name
            data['addresses'][i] = addr

        if isinstance(data.get('duplex'), Enum):
            data['duplex'] = data['duplex'].name.split('_')[-1]

        # Split the flags string
        data['flags'] = data.get('flags', '').split(',')

        return data



from enum import Enum


class DeviceType(Enum):
    """
    Constants used for the `device_type` attribute.

    Reference: https://github.com/OpenWonderLabs/SwitchBotAPI
    """

    BLIND_TILT = 'Blind Tilt'
    BOT = 'Bot'
    CEILING_LIGHT = 'Ceiling Light'
    CEILING_LIGHT_PRO = 'Ceiling Light Pro'
    COLOR_BULB = 'Color Bulb'
    CONTACT_SENSOR = 'Contact Sensor'
    CURTAIN = 'Curtain'
    HUMIDIFIER = 'Humidifier'
    KEYPAD = 'Keypad'
    KEYPAD_TOUCH = 'Keypad Touch'
    LOCK = 'Smart Lock'
    METER = 'Meter'
    METER_PLUS = 'Meter Plus'
    MOTION_SENSOR = 'Motion Sensor'
    PLUG = 'Plug'
    PLUG_MINI_JP = 'Plug Mini (JP)'
    PLUG_MINI_US = 'Plug Mini (US)'
    ROBOT_VACUUM_CLEANER_S1 = 'Robot Vacuum Cleaner S1'
    ROBOT_VACUUM_CLEANER_S1_PLUS = 'Robot Vacuum Cleaner S1 Plus'
    STRIP_LIGHT = 'Strip Light'



from typing import List

import requests

from .._model import TorrentSearchResult
from ._base import TorrentsCsvBaseProvider


class TorrentsCsvAPIProvider(TorrentsCsvBaseProvider):
    """
    Torrent that uses `Torrents.csv <https://torrents-csv.com/>`_ or any other
    `Torrents.csv API <https://torrents-csv.com/service>`_ instance to search
    for torrents.
    """

    def __init__(self, api_url: str, **kwargs):
        """
        :param api_url: Torrents.csv API base URL.
        """
        super().__init__(**kwargs)
        self.api_url = api_url

    def _search(
        self, query: str, *_, limit: int, page: int, **__
    ) -> List[TorrentSearchResult]:
        """
        Perform a search of torrents using the Torrent.csv API.

        :param query: Query string.
        :param limit: Number of results to return (default: 25).
        :param page: Page number (default: 1).
        """
        response = requests.get(
            f'{self.api_url}/search',
            params={
                'q': query,
                'size': limit,
                'page': page,
            },
            timeout=self._http_timeout,
        )

        response.raise_for_status()
        return [
            TorrentSearchResult(
                title=torrent.get('name', '[No Title]'),
                url=self._to_magnet(
                    info_hash=torrent.get('infohash'), torrent_name=torrent.get('name')
                ),
                size=torrent.get('size_bytes'),
                created_at=torrent.get('created_unix'),
                seeds=torrent.get('seeders'),
                peers=torrent.get('leechers'),
            )
            for torrent in response.json().get('torrents', [])
        ]


# vim:sw=4:ts=4:et:



from typing import List, Optional

import sounddevice as sd

from .._model import AudioDevice, DeviceType, StreamType


class DeviceManager:
    """
    The device manager is responsible for managing the virtual audio device
    abstractions exposed by the OS.

    For example, on a pure ALSA system virtual devices are usually mapped the
    physical audio devices available on the system.

    On a system that runs through PulseAudio or Jack, there may be a
    ``default`` virtual device whose sound card mappings may be managed by the
    audio server.
    """

    def __init__(
        self,
        input_device: Optional[DeviceType] = None,
        output_device: Optional[DeviceType] = None,
    ):
        """
        :param input_device: The default input device to use (by index or name).
        :param output_device: The default output device to use (by index or name).
        """
        self.input_device = (
            self.get_device(input_device, StreamType.INPUT)
            if input_device is not None
            else None
        )

        self.output_device = (
            self.get_device(output_device, StreamType.OUTPUT)
            if output_device is not None
            else None
        )

    def get_devices(
        self, type: Optional[StreamType] = None  # pylint: disable=redefined-builtin
    ) -> List[AudioDevice]:
        """
        Get available audio devices.

        :param type: The type of devices to filter (default: return all).
        """
        devices: List[dict] = sd.query_devices()  # type: ignore
        if type:
            devices = [dev for dev in devices if dev.get(f'max_{type.value}_channels')]

        return [
            AudioDevice(**{'index': idx, **info}) for idx, info in enumerate(devices)
        ]

    def get_device(
        self,
        device: Optional[DeviceType] = None,
        type: Optional[StreamType] = None,  # pylint: disable=redefined-builtin
    ) -> AudioDevice:
        """
        Search for a device.

        Either ``device`` or ``type`` have to be specified.

        :param device: The device to search for, either by index or name. If
            not specified, then the default device for the given type is
            returned.
        :param type: The type of the device to search.
        """
        assert device or type, 'Please specify either device or type'
        if device is None:
            if type == StreamType.INPUT and self.input_device is not None:
                return self.input_device
            if type == StreamType.OUTPUT and self.output_device is not None:
                return self.output_device

        all_devices: List[dict] = sd.query_devices()  # type: ignore
        assert all_devices, 'No sound devices found'

        try:
            dev: dict = sd.query_devices(
                kind=type.value if type else None, device=device  # type: ignore
            )
        except sd.PortAudioError as e:
            raise AssertionError(
                f'Could not get device for type={type} and device={device}: {e}',
                type,
                device,
                e,
            ) from e

        if not dev:
            if device:
                raise AssertionError(f'No such device: {device}')
            if type:
                raise AssertionError(f'No default device for type={type}')

        if dev.get('idx') is None:
            idx = next(
                iter(i for i, d in enumerate(all_devices) if d['name'] == dev['name']),
                None,
            )

            assert idx is not None, 'Could not infer the sound device index'
            dev['index'] = idx

        return AudioDevice(**dev)



import io
from abc import ABC

from PIL.Image import Image

from platypush.plugins.camera.model.writer import StreamWriter


class ImageStreamWriter(StreamWriter, ABC):
    """
    Write camera frames to a stream as single JPEG items.
    """

    @staticmethod
    def _encode(image: Image, encoding: str, **kwargs) -> bytes:
        with io.BytesIO() as buf:
            image.save(buf, format=encoding, **kwargs)
            return buf.getvalue()


class JPEGStreamWriter(ImageStreamWriter):
    """
    Write camera frames to a stream as single JPEG items.
    """
    mimetype = 'image/jpeg'

    def __init__(self, *args, quality: int = 90, **kwargs):
        super().__init__(*args, **kwargs)
        assert 0 < quality <= 100, 'JPEG quality should be between 0 and 100'
        self.quality = quality

    def encode(self, image: Image) -> bytes:
        return self._encode(image, 'jpeg', quality=self.quality)


class PNGStreamWriter(ImageStreamWriter):
    """
    Write camera frames to a stream as single PNG items.
    """
    mimetype = 'image/png'

    def encode(self, image: Image) -> bytes:
        return self._encode(image, 'png')


class BMPStreamWriter(ImageStreamWriter):
    """
    Write camera frames to a stream as single BMP items.
    """
    mimetype = 'image/bmp'

    def encode(self, image: Image) -> bytes:
        return self._encode(image, 'bmp')


class MJPEGStreamWriter(JPEGStreamWriter):
    """
    Write camera frames to a stream as an MJPEG feed.
    """
    mimetype = 'multipart/x-mixed-replace; boundary=frame'

    def encode(self, image: Image) -> bytes:
        return (b'--frame\r\n'
                b'Content-Type: image/jpeg\r\n\r\n' + super().encode(image) + b'\r\n')


# vim:sw=4:ts=4:et:



import argparse
import os
import re
import sys
import textwrap as tw
from typing import List, Optional

import httplib2
from oauth2client import client
from oauth2client import tools
from oauth2client.file import Storage

from platypush.config import Config

credentials_dir = os.path.join(Config.get_workdir(), "credentials", "google")
default_secrets_file = os.path.join(credentials_dir, "client_secret.json")
"""Default path for the Google API client secrets file"""


def _parse_scopes(*scopes: str) -> List[str]:
    return sorted(
        {
            t.split("/")[-1].strip()
            for scope in scopes
            for t in re.split(r"[\s,]", scope)
            if t
        }
    )


def get_credentials_filename(*scopes: str):
    parsed_scopes = _parse_scopes(*scopes)
    scope_name = "-".join([scope.split("/")[-1] for scope in parsed_scopes])
    os.makedirs(credentials_dir, exist_ok=True)
    matching_scope_file = next(
        iter(
            os.path.join(credentials_dir, scopes_file)
            for scopes_file in {
                os.path.basename(file)
                for file in os.listdir(credentials_dir)
                if file.endswith(".json")
            }
            if not set(parsed_scopes).difference(
                set(scopes_file.split(".json")[0].split("-"))
            )
        ),
        None,
    )

    if matching_scope_file:
        return matching_scope_file

    return os.path.join(credentials_dir, scope_name + ".json")


def get_credentials(scope: str, secrets_file: Optional[str] = None):
    scopes = _parse_scopes(scope)
    credentials_file = get_credentials_filename(*scopes)

    # If we don't have a credentials file for the required set of scopes, but we have a secrets file,
    # then try and generate the credentials file from the stored secrets.
    if (
        not os.path.isfile(credentials_file)
        and secrets_file
        and os.path.isfile(secrets_file)
        and (os.getenv("DISPLAY") or os.getenv("BROWSER"))
    ):
        args = []
        generate_credentials(secrets_file, scope, *args)

    assert os.path.isfile(credentials_file), tw.dedent(
        f"""
        Credentials file {credentials_file} not found. Generate it through:
            python -m platypush.plugins.google.credentials "{','.join(scopes)}" {
                secrets_file or '/path/to/client_secret.json'
            }
              [--auth_host_name AUTH_HOST_NAME]
              [--noauth_local_webserver]
              [--auth_host_port [AUTH_HOST_PORT [AUTH_HOST_PORT ...]]]
              [--logging_level [DEBUG,INFO,WARNING,ERROR,CRITICAL]]

        Specify --noauth_local_webserver if you're running this script on a headless machine.
        You will then get an authentication URL on the logs.
        Otherwise, the URL will be opened in the available browser.
        """
    )

    store = Storage(credentials_file)
    credentials = store.get()

    if not credentials or credentials.invalid:
        credentials.refresh(httplib2.Http())

    return credentials


def generate_credentials(client_secret_path: str, scope: str, *args: str):
    scopes = _parse_scopes(scope)
    credentials_file = get_credentials_filename(*scopes)
    store = Storage(credentials_file)
    scope = ' '.join(
        f'https://www.googleapis.com/auth/{scope}' for scope in _parse_scopes(scope)
    )

    flow = client.flow_from_clientsecrets(client_secret_path, scope)
    flow.user_agent = "Platypush"
    flow.access_type = "offline"  # type: ignore
    flags = argparse.ArgumentParser(parents=[tools.argparser]).parse_args(args)  # type: ignore
    tools.run_flow(flow, store, flags)
    print("Storing credentials to", credentials_file)
    print(
        "\nIf this is not the working directory of your Platypush instance, \n"
        "then move the generated credentials file to WORKDIR/credentials/google"
    )


def main():
    """
    Generates a Google API credentials file given client secret JSON and scopes.
    Usage::

        python -m platypush.plugins.google.credentials \
            [spaces/comma-separated list of scopes] \
            [client_secret.json location]

    """

    args = sys.argv[1:]
    scope = (
        args.pop(0) if args else input("Space/comma separated list of OAuth scopes: ")
    ).strip()

    if args:
        client_secret_path = args.pop(0)
    elif os.path.isfile(default_secrets_file):
        client_secret_path = default_secrets_file
    else:
        client_secret_path = input("Google credentials JSON file location: ")

    client_secret_path = os.path.abspath(os.path.expanduser(client_secret_path)).strip()
    generate_credentials(client_secret_path, scope, *args)


if __name__ == "__main__":
    main()

# vim:sw=4:ts=4:et:



import datetime
import json
from threading import Event
from typing import Optional, Dict, List, Union

import trello
from trello.board import Board, List as List_  # type: ignore
from trello.exceptions import ResourceUnavailable  # type: ignore
from websocket import WebSocketApp

from platypush.context import get_bus
from platypush.message.event.trello import (
    MoveCardEvent,
    NewCardEvent,
    ArchivedCardEvent,
    UnarchivedCardEvent,
)
from platypush.plugins import RunnablePlugin, action
from platypush.schemas.trello import (
    TrelloBoardSchema,
    TrelloCardSchema,
    TrelloListSchema,
    TrelloMemberSchema,
)

from ._model import (
    TrelloBoard,
    TrelloCard,
    TrelloAttachment,
    TrelloPreview,
    TrelloChecklist,
    TrelloChecklistItem,
    TrelloUser,
    TrelloComment,
    TrelloLabel,
    TrelloList,
    TrelloMember,
)


class TrelloPlugin(RunnablePlugin):
    """
    Trello integration.

    You'll need a Trello API key. You can get it `here
    <https://trello.com/app-key>`_.

    You'll also need an auth token if you want to view/change private
    resources. You can generate a permanent token linked to your account on
    ``https://trello.com/1/connect?key=<KEY>&name=platypush&response_type=token&expiration=never&scope=read,write``.

    Also, polling of events requires you to follow a separate procedure to
    retrieve the Websocket tokens, since Trello uses a different (and
    undocumented) authorization mechanism:

        1. Open https://trello.com in your browser.
        2. Open the developer tools (F12).
        3. Go to the Cookies tab.
        4. Copy the value of the ``cloud.session.token`` cookie.

    """

    _websocket_url_base = 'wss://trello.com/1/Session/socket?clientVersion=build-194674'

    def __init__(
        self,
        api_key: str,
        api_secret: Optional[str] = None,
        token: Optional[str] = None,
        cloud_session_token: Optional[str] = None,
        boards: Optional[List[str]] = None,
        **kwargs,
    ):
        """
        :param api_key: Trello API key. You can get it `here
            <https://trello.com/app-key>`_.
        :param api_secret: Trello API secret. You can get it `here
            <https://trello.com/app-key>`_.
        :param token: Trello token. It is required if you want to access or
            modify private resources. You can get a permanent token on
            ``https://trello.com/1/connect?key=<KEY>&name=platypush&response_type=token&expiration=never&scope=read,write``.
        :param cloud_session_token: Cloud session token. It is required
            if you want to monitor your boards for changes. See
            :class:`platypush.plugins.trello.TrelloPlugin` for the procedure to
            retrieve it.
        :param boards: List of boards to subscribe, by ID or name. If
            specified, then the plugin will listen for changes only on these boards.
        """

        super().__init__(**kwargs)
        self.api_key = api_key
        self.api_secret = api_secret
        self.token = token
        self.cloud_session_token = cloud_session_token
        self._client = None
        self._req_id = 0
        self._boards_by_id = {}
        self._boards_by_name = {}
        self._monitored_boards = boards
        self.url = None

        if token:
            self.url = self._websocket_url_base

        self._connected = Event()
        self._items = {}
        self._event_handled = False
        self._ws: Optional[WebSocketApp] = None

    def _get_client(self) -> trello.TrelloClient:  # type: ignore
        if not self._client:
            self._client = trello.TrelloClient(  # type: ignore
                api_key=self.api_key,
                api_secret=self.api_secret,
                token=self.token,
            )

        return self._client

    def _get_board(self, board: str) -> Board:
        client = self._get_client()
        try:
            return client.get_board(board)
        except ResourceUnavailable:
            boards = [b for b in client.list_boards() if b.name == board]
            assert boards, f'No such board: {board}'
            return boards[0]

    def _get_boards(
        self, all: bool = False  # pylint: disable=redefined-builtin
    ) -> List[Board]:
        client = self._get_client()
        return client.list_boards(board_filter='all' if all else 'open')

    @action
    def get_boards(self, all: bool = False):  # pylint: disable=redefined-builtin
        """
        Get the list of boards.

        :param all: If True, return all the boards included those that have
            been closed/archived/deleted. Otherwise, only return open/active
            boards (default: False).
        :return: .. schema:: trello.TrelloBoardSchema(many=True)
        """
        return TrelloBoardSchema().dump(
            [
                TrelloBoard(
                    id=b.id,
                    name=b.name,
                    description=b.description,
                    url=b.url,
                    date_last_activity=b.date_last_activity,
                    closed=b.closed,
                )
                for b in self._get_boards(all=all)
            ],
            many=True,
        )

    @action
    def get_board(self, board: str):
        """
        Get the info about a board.

        :param board: Board ID or name
        :return: .. schema:: trello.TrelloBoardSchema
        """

        b = self._get_board(board)
        return TrelloBoardSchema().dump(
            TrelloBoard(
                id=b.id,
                name=b.name,
                url=b.url,
                closed=b.closed,
                description=b.description,
                date_last_activity=b.date_last_activity,
                lists=[
                    TrelloList(
                        id=ll.id,
                        name=ll.name,
                        closed=ll.closed,
                        subscribed=ll.subscribed,
                    )
                    for ll in b.list_lists()
                ],
            )
        )

    @action
    def open_board(self, board: str):
        """
        Re-open/un-archive a board.

        :param board: Board ID or name
        """
        self._get_board(board).open()

    @action
    def close_board(self, board: str):
        """
        Close/archive a board.

        :param board: Board ID or name
        """
        self._get_board(board).close()

    @action
    def set_board_name(self, board: str, name: str):
        """
        Change the name of a board.

        :param board: Board ID or name.
        :param name: New name.
        """
        self._get_board(board).set_name(name)
        return self.get_board(name)

    @action
    def set_board_description(self, board: str, description: str):
        """
        Change the description of a board.

        :param board: Board ID or name.
        :param description: New description.
        """
        self._get_board(board).set_description(description)
        return self.get_board(description)

    @action
    def create_label(self, board: str, name: str, color: Optional[str] = None):
        """
        Add a label to a board.

        :param board: Board ID or name
        :param name: Label name
        :param color: Optional HTML color
        """
        self._get_board(board).add_label(name=name, color=color)

    @action
    def delete_label(self, board: str, label: str):
        """
        Delete a label from a board.

        :param board: Board ID or name
        :param label: Label ID or name
        """

        b = self._get_board(board)

        try:
            b.delete_label(label)
        except ResourceUnavailable:
            label_ = next(iter(ll for ll in b.get_labels() if ll.name == label), None)
            assert label_, f'No such label: {label}'
            label = label_.id
            b.delete_label(label)

    @action
    def add_member(self, board: str, member_id: str, member_type: str = 'normal'):
        """
        Add a member to a board.

        :param board: Board ID or name.
        :param member_id: Member ID to add.
        :param member_type: Member type - can be 'normal' or 'admin' (default: 'normal').
        """
        self._get_board(board).add_member(member_id, member_type=member_type)

    @action
    def remove_member(self, board: str, member_id: str):
        """
        Remove a member from a board.

        :param board: Board ID or name.
        :param member_id: Member ID to remove.
        """
        self._get_board(board).remove_member(member_id)

    def _get_members(self, board: str, only_admin: bool = False):
        b = self._get_board(board)
        members = b.admin_members() if only_admin else b.get_members()

        return TrelloMemberSchema().dump(
            [
                TrelloMember(
                    id=m.id,
                    fullname=m.full_name,
                    bio=m.bio,
                    url=m.url,
                    username=m.username,
                    initials=m.initials,
                    member_type=getattr(m, 'member_type', None),
                )
                for m in members
            ],
            many=True,
        )

    @action
    def get_members(self, board: str):
        """
        Get the list of all the members of a board.

        :param board: Board ID or name.
        :return: .. schema:: trello.TrelloMemberSchema(many=True)
        """
        return self._get_members(board, only_admin=False)

    @action
    def get_admin_members(self, board: str):
        """
        Get the list of the admin members of a board.

        :param board: Board ID or name.
        :return: .. schema:: trello.TrelloMemberSchema(many=True)
        """
        return self._get_members(board, only_admin=True)

    @action
    def get_lists(
        self, board: str, all: bool = False  # pylint: disable=redefined-builtin
    ):
        """
        Get the list of lists on a board.

        :param board: Board ID or name
        :param all: If True, return all the lists, included those that have been closed/archived/deleted. Otherwise,
            only return open/active lists (default: False).
        :return: .. schema:: trello.TrelloListSchema(many=True)
        """
        return TrelloListSchema().dump(
            [
                TrelloList(
                    id=ll.id, name=ll.name, closed=ll.closed, subscribed=ll.subscribed
                )
                for ll in self._get_board(board).list_lists('all' if all else 'open')
            ],
            many=True,
        )

    @action
    def add_list(self, board: str, name: str, pos: Optional[int] = None):
        """
        Add a list to a board.

        :param board: Board ID or name
        :param name: List name
        :param pos: Optional position (default: last)
        """
        self._get_board(board).add_list(name=name, pos=pos)

    def _get_list(
        self, board: str, list: str  # pylint: disable=redefined-builtin
    ) -> List_:
        b = self._get_board(board)

        try:
            return b.get_list(list)
        except ResourceUnavailable:
            lists = [ll for ll in b.list_lists() if ll.name == list]
            assert lists, f'No such list: {list}'
            return lists[0]

    @action
    def set_list_name(
        self, board: str, list: str, name: str  # pylint: disable=redefined-builtin
    ):
        """
        Change the name of a board list.

        :param board: Board ID or name
        :param list: List ID or name
        :param name: New name
        """
        self._get_list(board, list).set_name(name)

    @action
    def list_subscribe(
        self, board: str, list: str  # pylint: disable=redefined-builtin
    ):
        """
        Subscribe to a list.

        :param board: Board ID or name
        :param list: List ID or name
        """
        self._get_list(board, list).subscribe()

    @action
    def list_unsubscribe(
        self, board: str, list: str  # pylint: disable=redefined-builtin
    ):
        """
        Unsubscribe from a list.

        :param board: Board ID or name
        :param list: List ID or name
        """
        self._get_list(board, list).unsubscribe()

    @action
    def archive_all_cards(
        self, board: str, list: str  # pylint: disable=redefined-builtin
    ):
        """
        Archive all the cards on a list.

        :param board: Board ID or name
        :param list: List ID or name
        """
        self._get_list(board, list).archive_all_cards()

    @action
    def move_all_cards(self, board: str, src: str, dest: str):
        """
        Move all the cards from a list to another.

        :param board: Board ID or name
        :param src: Source list
        :param dest: Target list
        """
        src_list = self._get_list(board, src)
        dest_list = self._get_list(board, dest)
        src_list.move_all_cards(dest_list.id)

    @action
    def open_list(self, board: str, list: str):  # pylint: disable=redefined-builtin
        """
        Open/un-archive a list.

        :param board: Board ID or name
        :param list: List ID or name
        """
        self._get_list(board, list).open()

    @action
    def close_list(self, board: str, list: str):  # pylint: disable=redefined-builtin
        """
        Close/archive a list.

        :param board: Board ID or name
        :param list: List ID or name
        """
        self._get_list(board, list).close()

    @action
    def move_list(
        self, board: str, list: str, position: int  # pylint: disable=redefined-builtin
    ):
        """
        Move a list to another position.

        :param board: Board ID or name
        :param list: List ID or name
        :param position: New position index
        """
        self._get_list(board, list).move(position)

    @action
    def add_card(
        self,
        board: str,
        list: str,  # pylint: disable=redefined-builtin
        name: str,
        description: Optional[str] = None,
        position: Optional[int] = None,
        labels: Optional[List[str]] = None,
        due: Optional[Union[str, datetime.datetime]] = None,
        source: Optional[str] = None,
        assign: Optional[List[str]] = None,
    ):
        """
        Add a card to a list.

        :param board: Board ID or name
        :param list: List ID or name
        :param name: Card name
        :param description: Card description
        :param position: Card position index
        :param labels: List of labels
        :param due: Due date (``datetime.datetime`` object or ISO-format string)
        :param source: Card ID to clone from
        :param assign: List of assignee member IDs
        :return: .. schema:: trello.TrelloCardSchema
        """
        list_ = self._get_list(board, list)

        if labels:
            labels = [
                ll
                for ll in list_.board.get_labels()
                if ll.id in labels or ll.name in labels
            ]

        card = list_.add_card(
            name=name,
            desc=description,
            labels=labels,
            due=due,
            source=source,
            position=position,
            assign=assign,
        )

        return TrelloCardSchema().dump(
            TrelloCard(
                id=card.id,
                name=card.name,
                url=card.url,
                closed=card.closed,
                board=TrelloBoard(
                    id=list_.board.id,
                    name=list_.board.name,
                    url=list_.board.url,
                    closed=list_.board.closed,
                    description=list_.board.description,
                    date_last_activity=list_.board.date_last_activity,
                ),
                is_due_complete=card.is_due_complete,
                list=None,
                comments=[],
                labels=[
                    TrelloLabel(id=lb.id, name=lb.name, color=lb.color)
                    for lb in (card.labels or [])
                ],
                description=card.description,
                due_date=card.due_date,
                latest_card_move_date=card.latestCardMove_date,
                date_last_activity=card.date_last_activity,
            )
        )

    @action
    def delete_card(self, card_id: str):
        """
        Permanently delete a card.

        :param card_id: Card ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.delete()

    @action
    def open_card(self, card_id: str):
        """
        Open/un-archive a card.

        :param card_id: Card ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.set_closed(False)

    @action
    def close_card(self, card_id: str):
        """
        Close/archive a card.

        :param card_id: Card ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.set_closed(True)

    @action
    def add_checklist(
        self,
        card_id: str,
        title: str,
        items: List[str],
        states: Optional[List[bool]] = None,
    ):
        """
        Add a checklist to a card.

        :param card_id: Card ID
        :param title: Checklist title
        :param items: List of items in the checklist
        :param states: State of each item, True for checked, False for unchecked
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.add_checklist(title, items, states)

    @action
    def add_label(self, card_id: str, label: str):
        """
        Add a label to a card.

        :param card_id: Card ID
        :param label: Label name
        """
        client = self._get_client()
        card = client.get_card(card_id)

        labels = [ll for ll in card.board.get_labels() if ll.name == label]

        assert labels, f'No such label: {label}'
        label = labels[0]
        card.add_label(label)

    @action
    def remove_label(self, card_id: str, label: str):
        """
        Remove a label from a card.

        :param card_id: Card ID
        :param label: Label name
        """
        client = self._get_client()
        card = client.get_card(card_id)

        labels = [ll for ll in card.board.get_labels() if ll.name == label]

        assert labels, f'No such label: {label}'
        label = labels[0]
        card.remove_label(label)

    @action
    def assign_card(self, card_id: str, member_id: str):
        """
        Assign a card.

        :param card_id: Card ID
        :param member_id: Member ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.assign(member_id)

    @action
    def unassign_card(self, card_id: str, member_id: str):
        """
        Un-assign a card.

        :param card_id: Card ID
        :param member_id: Member ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.unassign(member_id)

    @action
    def attach_card(
        self,
        card_id: str,
        name: Optional[str] = None,
        mime_type: Optional[str] = None,
        file: Optional[str] = None,
        url: Optional[str] = None,
    ):
        """
        Add an attachment to a card. It can be either a local file or a remote URL.

        :param card_id: Card ID
        :param name: File name
        :param mime_type: MIME type
        :param file: Path to the file
        :param url: URL to the file
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.attach(name=name, mimeType=mime_type, file=file, url=url)

    @action
    def remove_attachment(self, card_id: str, attachment_id: str):
        """
        Remove an attachment from a card.

        :param card_id: Card ID
        :param attachment_id: Attachment ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.remove_attachment(attachment_id)

    @action
    def change_card_board(
        self,
        card_id: str,
        board: str,
        list: Optional[str] = None,  # pylint: disable=redefined-builtin
    ):
        """
        Move a card to a new board.

        :param card_id: Card ID
        :param board: New board ID or name
        :param list: Optional target list ID or name
        """
        client = self._get_client()
        card = client.get_card(card_id)
        board_id = self._get_board(board).id

        list_id = None
        if list:
            list_id = self._get_list(board_id, list).id

        card.change_board(board_id, list_id)

    @action
    def change_card_list(
        self, card_id: str, list: str  # pylint: disable=redefined-builtin
    ):
        """
        Move a card to a new list.

        :param card_id: Card ID
        :param list: List ID or name
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.change_list(self._get_list(card.board.id, list).id)

    @action
    def change_card_pos(self, card_id: str, position: int):
        """
        Move a card to a new position.

        :param card_id: Card ID
        :param position: New position index
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.change_pos(position)

    @action
    def comment_card(self, card_id: str, text: str):
        """
        Add a comment to a card.

        :param card_id: Card ID
        :param text: Comment text
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.comment(text)

    @action
    def update_comment(self, card_id: str, comment_id: str, text: str):
        """
        Update the content of a comment.

        :param card_id: Card ID
        :param comment_id: Comment ID
        :param text: New comment text
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.update_comment(comment_id, text)

    @action
    def delete_comment(self, card_id: str, comment_id: str):
        """
        Delete a comment.

        :param card_id: Card ID
        :param comment_id: Comment ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.delete_comment(comment_id)

    @action
    def set_card_name(self, card_id: str, name: str):
        """
        Change the name of a card.

        :param card_id: Card ID
        :param name: New name
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.set_name(name)

    @action
    def set_card_description(self, card_id: str, description: str):
        """
        Change the description of a card.

        :param card_id: Card ID
        :param description: New description
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.set_description(description)

    @action
    def add_card_member(self, card_id: str, member_id: str):
        """
        Add a member to a card.

        :param card_id: Card ID
        :param member_id: Member ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.add_member(member_id)

    @action
    def remove_card_member(self, card_id: str, member_id: str):
        """
        Remove a member from a card.

        :param card_id: Card ID
        :param member_id: Member ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.remove_member(member_id)

    @action
    def set_card_due(self, card_id: str, due: Union[str, datetime.datetime]):
        """
        Set the due date for a card.

        :param card_id: Card ID
        :param due: Due date, as a datetime.datetime object or an ISO string
        """
        client = self._get_client()
        card = client.get_card(card_id)

        if isinstance(due, str):
            due = datetime.datetime.fromisoformat(due)

        card.set_due(due)

    @action
    def remove_card_due(self, card_id: str):
        """
        Remove the due date from a card.

        :param card_id: Card ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.remove_due()

    @action
    def set_card_due_complete(self, card_id: str):
        """
        Set the due date of a card as completed.

        :param card_id: Card ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.set_due_complete()

    @action
    def remove_card_due_complete(self, card_id: str):
        """
        Remove the due complete flag from a card.

        :param card_id: Card ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.remove_due_complete()

    @action
    def card_subscribe(self, card_id: str):
        """
        Subscribe to a card.
        :param card_id: Card ID
        """
        client = self._get_client()
        card = client.get_card(card_id)
        card.subscribe()

    @action
    def get_cards(
        self,
        board: str,
        list: Optional[str] = None,  # pylint: disable=redefined-builtin
        all: bool = False,  # pylint: disable=redefined-builtin
    ):
        """
        Get the list of cards on a board.

        :param board: Board ID or name
        :param list: List ID or name. If set then the method will only return the cards found on that list
            (default: None)
        :param all: If True, return all the cards included those that have been closed/archived/deleted. Otherwise,
            only return open/active cards (default: False).
        :return: .. schema:: trello.TrelloCardSchema(many=True)
        """

        b = self._get_board(board)
        lists: Dict[str, TrelloList] = {
            ll.id: TrelloList(
                id=ll.id, name=ll.name, closed=ll.closed, subscribed=ll.subscribed
            )
            for ll in b.list_lists()
        }

        list_id = None

        if list:
            if list in lists:
                list_id = list
            else:
                ll = next(
                    iter(
                        l1 for l1 in lists.values() if l1.name == list  # type: ignore
                    ),
                    None,
                )
                assert ll, f'No such list ID/name: {list}'
                list_id = ll.id  # type: ignore

        return TrelloCardSchema().dump(
            [
                TrelloCard(
                    id=c.id,
                    name=c.name,
                    url=c.url,
                    closed=c.closed,
                    list=lists.get(c.list_id),
                    board=TrelloBoard(
                        id=c.board.id,
                        name=c.board.name,
                        url=c.board.url,
                        closed=c.board.closed,
                        description=c.board.description,
                        date_last_activity=c.board.date_last_activity,
                    ),
                    attachments=[
                        TrelloAttachment(
                            id=a.get('id'),
                            url=a.get('url'),
                            size=a.get('bytes'),
                            date=a.get('date'),
                            edge_color=a.get('edgeColor'),
                            member_id=a.get('idMember'),
                            is_upload=a.get('isUpload'),
                            name=a.get('name'),
                            previews=[
                                TrelloPreview(
                                    id=p.get('id'),
                                    scaled=p.get('scaled'),
                                    url=p.get('url'),
                                    size=p.get('bytes'),
                                    height=p.get('height'),
                                    width=p.get('width'),
                                )
                                for p in a.get('previews', [])
                            ],
                            mime_type=a.get('mimeType'),
                        )
                        for a in c.attachments
                    ],
                    checklists=[
                        TrelloChecklist(
                            id=ch.id,
                            name=ch.name,
                            items=[
                                TrelloChecklistItem(
                                    id=i.get('id'),
                                    name=i.get('name'),
                                    checked=i.get('checked'),
                                )
                                for i in ch.items
                            ],
                        )
                        for ch in c.checklists
                    ],
                    comments=[
                        TrelloComment(
                            id=co.get('id'),
                            text=co.get('data', {}).get('text'),
                            type=co.get('type'),
                            date=co.get('date'),
                            creator=TrelloUser(
                                id=co.get('memberCreator', {}).get('id'),
                                username=co.get('memberCreator', {}).get('username'),
                                fullname=co.get('memberCreator', {}).get('fullName'),
                                initials=co.get('memberCreator', {}).get('initials'),
                                avatar_url=co.get('memberCreator', {}).get('avatarUrl'),
                            ),
                        )
                        for co in c.comments
                    ],
                    labels=[
                        TrelloLabel(id=lb.id, name=lb.name, color=lb.color)
                        for lb in (c.labels or [])
                    ],
                    is_due_complete=c.is_due_complete,
                    due_date=c.due_date,
                    description=c.description,
                    latest_card_move_date=c.latestCardMove_date,
                    date_last_activity=c.date_last_activity,
                )
                for c in b.all_cards()
                if ((all or not c.closed) and (not list or c.list_id == list_id))
            ],
            many=True,
        )

    def _initialize_connection(self, ws: WebSocketApp):
        boards = [
            b
            for b in (self._get_boards() or [])
            if not self._monitored_boards
            or b.id in self._monitored_boards
            or b.name in self._monitored_boards
        ]

        for b in boards:
            self._boards_by_id[b.id] = b
            self._boards_by_name[b.name] = b

        for board_id in self._boards_by_id.keys():
            self._send(
                ws,
                {
                    'type': 'subscribe',
                    'modelType': 'Board',
                    'idModel': board_id,
                    'tags': ['clientActions', 'updates'],
                    'invitationTokens': [],
                },
            )

        self.logger.info('Trello boards subscribed')

    def _on_msg(self, *args):  # pylint: disable=too-many-return-statements
        if len(args) < 2:
            self.logger.warning(
                'Missing websocket argument - make sure that you are using '
                'a version of websocket-client < 0.53.0 or >= 0.58.0'
            )
            return

        ws, msg = args[:2]
        if not msg:
            # Reply back with an empty message when the server sends an empty message
            ws.send('')
            return

        try:
            msg = json.loads(msg)
        except Exception as e:
            self.logger.warning(
                'Received invalid JSON message from Trello: %s: %s', msg, e
            )
            return

        if 'error' in msg:
            self.logger.warning('Trello error: %s', msg['error'])
            return

        if msg.get('reqid') == 0:
            self.logger.debug('Ping response received, subscribing boards')
            self._initialize_connection(ws)
            return

        notify = msg.get('notify')
        if not notify:
            return

        if notify['event'] != 'updateModels' or notify['typeName'] != 'Action':
            return

        for delta in notify['deltas']:
            args = {
                'card_id': delta['data']['card']['id'],
                'card_name': delta['data']['card']['name'],
                'list_id': (
                    delta['data'].get('list') or delta['data'].get('listAfter', {})
                ).get('id'),
                'list_name': (
                    delta['data'].get('list') or delta['data'].get('listAfter', {})
                ).get('name'),
                'board_id': delta['data']['board']['id'],
                'board_name': delta['data']['board']['name'],
                'closed': delta.get('closed'),
                'member_id': delta['memberCreator']['id'],
                'member_username': delta['memberCreator']['username'],
                'member_fullname': delta['memberCreator']['fullName'],
                'date': delta['date'],
            }

            if delta.get('type') == 'createCard':
                get_bus().post(NewCardEvent(**args))
            elif delta.get('type') == 'updateCard':
                if 'listBefore' in delta['data']:
                    args.update(
                        {
                            'old_list_id': delta['data']['listBefore']['id'],
                            'old_list_name': delta['data']['listBefore']['name'],
                        }
                    )

                    get_bus().post(MoveCardEvent(**args))
                elif 'closed' in delta['data'].get('old', {}):
                    cls = (
                        UnarchivedCardEvent
                        if delta['data']['old']['closed']
                        else ArchivedCardEvent
                    )
                    get_bus().post(cls(**args))

    def _on_error(self, *args):
        error = args[1] if len(args) > 1 else args[0]
        self.logger.warning('Trello websocket error: %s', error)

    def _on_close(self, *_):
        self.logger.warning('Trello websocket connection closed')
        self._connected.clear()
        self._req_id = 0

        while not self.should_stop():
            try:
                self._connect(reconnect=True)
                self._connected.wait(timeout=10)
                break
            except TimeoutError:
                continue

    def _on_open(self, *args):
        ws = args[0] if args else None
        self._connected.set()
        if ws:
            self._send(ws, {'type': 'ping'})
        self.logger.info('Trello websocket connected')

    def _send(self, ws: WebSocketApp, msg: dict):
        msg['reqid'] = self._req_id
        ws.send(json.dumps(msg))
        self._req_id += 1

    def _connect(self, reconnect: bool = False) -> WebSocketApp:
        assert self.url, 'Trello websocket URL not set'
        if reconnect:
            self.stop()

        if not self._ws:
            self._ws = WebSocketApp(
                self.url,
                header={
                    'Cookie': (
                        f'token={self.token}; '
                        f'cloud.session.token={self.cloud_session_token}'
                    )
                },
                on_open=self._on_open,
                on_message=self._on_msg,
                on_error=self._on_error,
                on_close=self._on_close,
            )

        return self._ws

    def main(self):
        if not self.url:
            self.logger.info(
                "token/cloud_session_token not set: your Trello boards won't be monitored for changes"
            )
            self.wait_stop()
        else:
            ws = self._connect()
            ws.run_forever()

    def stop(self):
        if self._ws:
            self._ws.close()
            self._ws = None


# vim:sw=4:ts=4:et:



import os
from dataclasses import dataclass
from datetime import datetime as dt
from enum import Enum
from threading import RLock
from typing import IO, Iterable, List, Optional

import requests

from platypush.plugins import Plugin, action


class ContextEntryRole(Enum):
    """
    Roles for context entries.
    """

    ASSISTANT = "assistant"
    SYSTEM = "system"
    USER = "user"


@dataclass
class ContextEntry:
    """
    A context entry.
    """

    timestamp: dt
    role: ContextEntryRole
    content: str

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            timestamp=dt.fromisoformat(data.get("timestamp", dt.now().isoformat())),
            role=ContextEntryRole(data["role"]),
            content=data["content"],
        )

    def to_dict(self):
        return {
            "role": self.role.value,
            "content": self.content,
        }


class OpenaiPlugin(Plugin):
    """
    Plugin to interact with OpenAI services.

    Currently supported:

        - :meth:`get_response`: Get a response to a prompt/question using the
          GPT API. It supports custom contexts and environment settings.

        - :meth:`transcribe`: Perform speech-to-text on an audio file. This API
          is also leveraged by the
          :class:`platypush.plugins.assistant.openai.OpenaiPlugin` to provide a
          full-fledged voice assistant.

        - Through the :class:`platypush.plugins.tts.openai.OpenaiPlugin` plugin,
          text-to-speech is also supported.

    Contexts
    --------

    The plugin also supports the implementation of custom assistant
    *contexts*/environment.

    Contexts can be used to:

        - Customize the model's behavior based on a set of inputs - going from
          a generic "*You are a helpful assistant*" to a more specific "*You
          are a Star Trek fan*", or "*You are a 16th century noble lady who
          talks in Shakespearean English to her peers*".
        - Pre-configure the model with a set of previous interactions in order
          to either pre-load information that we expect the model to remember,
          or to provide a set of previous interactions that the model can use
          to generate responses that are consistent with the conversation
          history.

    The plugin provides two types of contexts:

        - **Default context**: This is a set of context entries that are
          provided at plugin initialization and that will be used to initialize
          the model with a configuration or set of previous interactions that
          will be remembered when generating all responses.

        - **Runtime context**: This is a set of context entries that can be
          passed at runtime at :meth:`.get_response`. All the interactions
          (both user prompts and assistant responses) that are processed
          through :meth:`.get_response` will also be added to the runtime
          context, and remembered for the next ``context_expiry`` seconds. This
          allows you to generate responses that are consistent with the recent
          conversation history.

    Each context entry is a dictionary with the following keys:

        - ``role``: The role of the message. Can be one of:
            - ``system``: A system message provided to the model to set
              up its initial state - e.g. "you are a helpful
              assistant".
            - ``user``: A user message, as provided by a previous (real
              or synthetic) user interaction.
            - ``assistant``: An assistant message, as provided by a
              previous (real or synthetic) assistant response.
        - ``content``: The content of the message.

    An example of context:

      .. code-block:: yaml

        context:
            - role: system
              content: >
                  You are a 16th century noble lady who talks in
                  Shakespearean English to her peers.
            - role: user
              content: What is a telephone?
            - role: assistant
              content: >
                  Pray tell, noble companion, a telephone is a device
                  of modern innovation that doth permit one to speak
                  with a distant acquaintance by means of magical pink
                  waves that do carry the sound of thine voice to the
                  ear of the listener.

    Given such context, if you call :meth:`.get_response` with a
    prompt such as "*How does it work?*", the model may generate a
    response such as "*Fair lady, to use a telephone, thou must first
    lift the receiver and place it to thine ear. Then, thou must speak
    into the mouthpiece as though conversing with a companion in
    another room. The magical pink waves shall carry thy words to the
    recipient, who shall hear them on their own device. 'Tis a wondrous
    invention indeed!*".

    Note that the model will remember the previous interactions and
    also generate responses, so you can ask it direct questions such as "How
    does it work" while remembering what "it" is likely to mean. And it'll
    provide responses which are in the same style initialized through the
    ``system`` context.
    """

    def __init__(
        self,
        api_key: Optional[str],
        model: str = "gpt-3.5-turbo",
        timeout: float = 30,
        context: Optional[Iterable[dict]] = None,
        context_expiry: Optional[float] = 600,
        max_tokens: int = 500,
        **kwargs,
    ):
        """
        :param api_key: OpenAI API key. If not set, it will be read from the
            ``OPENAI_API_KEY`` environment variable.
        :param model: The model to use. Default: ``gpt-3.5-turbo``.
        :param timeout: Default timeout for API requests (default: 30 seconds).
        :param max_tokens: Maximum number of tokens to generate in the response
            (default: 500).
        :param context: Default context to use for completions, as a list of
            dictionaries with ``role`` and ``content`` keys. Default: None.
        :param context_expiry: Default expiry time for the context in seconds.
            After this time since the last interaction, the context will be
            cleared.

            This means that any follow-up interactions happening within the
            expiry window will remember the past prompts, but any interaction
            that happens after the expiry window (calculated from the time of
            the last interaction) will start fresh.

            Note that ``context_expiry`` is only applied to the runtime
            context. The default context will never expire unless it's removed
            from the plugin configuration.

            Set to 0 to disable context expiry - i.e. all messages stay in the
            context until the plugin is restarted or the context is cleared
            explicitly via :meth:`.clear_context`. Default: 600 seconds (10
            minutes).
        """
        super().__init__(**kwargs)
        api_key = api_key or os.getenv('OPENAI_API_KEY')
        assert api_key, 'OpenAI API key not provided'

        self._api_key = api_key
        self._context_lock = RLock()
        self._runtime_context: List[ContextEntry] = []
        self._default_context = [
            ContextEntry.from_dict(entries) for entries in (context or [])
        ]

        self.max_tokens = max_tokens
        self.context_expiry = context_expiry
        self.model = model
        self.timeout = timeout

    def _rotate_context(self):
        """
        Rotate the context by removing any entries older than the configured
        ``context_expiry``.
        """
        if not self.context_expiry:
            return

        with self._context_lock:
            now = dt.now()
            self._runtime_context = [
                entry
                for entry in self._runtime_context
                if (now - entry.timestamp).total_seconds() < self.context_expiry
            ]

    @action
    def get_response(
        self,
        prompt: str,
        model: Optional[str] = None,
        context: Optional[Iterable[dict]] = None,
        timeout: Optional[float] = None,
        max_tokens: Optional[int] = None,
    ) -> Optional[str]:
        """
        Get completions for a given prompt using ChatGPT.

        :param prompt: The prompt/question to complete/answer.
        :param model: Override the default model to use.
        :param context: Extend the default context with these extra messages.
        :param max_tokens: Override the default maximum number of tokens to
            generate in the response.
        :param timeout: Override the default timeout for the API request.
        :return: The completion for the prompt - or, better, the message
            associted to the highest scoring completion choice.
        """
        self._rotate_context()
        context = [
            *(context or []),
            {
                "role": "user",
                "content": prompt,
            },
        ]

        resp = requests.post(
            "https://api.openai.com/v1/chat/completions",
            timeout=timeout or self.timeout,
            headers={
                "Authorization": f"Bearer {self._api_key}",
                "Content-Type": "application/json",
            },
            json={
                "model": model or self.model,
                "messages": [
                    *(
                        entry.to_dict()
                        for entry in (
                            *(self._default_context or []),
                            *self._runtime_context,
                        )
                    ),
                    *context,
                ],
                "max_tokens": max_tokens or self.max_tokens,
            },
        )

        resp.raise_for_status()
        self._update_context(*context)
        choices = resp.json()["choices"]
        self.logger.debug("OpenAI response: %s", resp.json())

        if not choices:
            return None

        msg = choices[0]["message"]
        self._update_context(msg)
        return msg["content"]

    def _process_transcribe_response(self, resp: requests.Response) -> str:
        rs_json = None

        try:
            rs_json = resp.json()
        except Exception:
            pass

        self.logger.debug("OpenAI response: %s", rs_json)
        resp.raise_for_status()
        return (rs_json or {}).get("text", "")

    def transcribe_file(
        self,
        f: IO,
        model: Optional[str] = 'whisper-1',
        timeout: Optional[float] = None,
    ) -> str:
        resp = requests.post(
            "https://api.openai.com/v1/audio/transcriptions",
            timeout=timeout or self.timeout,
            headers={
                "Authorization": f"Bearer {self._api_key}",
            },
            files={
                "file": f,
            },
            data={
                "model": model or self.model,
            },
        )

        return self._process_transcribe_response(resp)

    def transcribe_raw(
        self,
        audio: bytes,
        extension: str,
        model: Optional[str] = 'whisper-1',
        timeout: Optional[float] = None,
    ) -> str:
        resp = requests.post(
            "https://api.openai.com/v1/audio/transcriptions",
            timeout=timeout or self.timeout,
            headers={
                "Authorization": f"Bearer {self._api_key}",
            },
            files={
                "file": (f"audio.{extension}", audio),
            },
            data={
                "model": model or self.model,
            },
        )

        return self._process_transcribe_response(resp)

    @action
    def transcribe(
        self,
        audio: str,
        model: Optional[str] = 'whisper-1',
        timeout: Optional[float] = None,
    ) -> str:
        """
        Perform speech-to-text on an audio file.

        :param audio: The audio file to transcribe.
        :param model: The model to use for speech-to-text. Default:
            ``whisper-1``. If not set, the configured default model will be
            used.
        :param timeout: Timeout for the API request. If not set, the default
            timeout will be used.
        :return: The transcribed text.
        """
        with open(os.path.expanduser(audio), "rb") as f:
            return self.transcribe_file(f, model=model, timeout=timeout)

    def _update_context(self, *entries: dict):
        """
        Update the context with a new entry.
        """
        with self._context_lock:
            for entry in entries:
                self._runtime_context.append(ContextEntry.from_dict(entry))
            self._rotate_context()

    @action
    def clear_context(self):
        """
        Clear the runtime context.
        """
        with self._context_lock:
            self._runtime_context = []



from abc import ABC
from typing import Optional, Tuple, Union
from platypush.message.event import Event


class SoundEvent(Event, ABC):
    """Base class for sound events"""

    def __init__(
        self, *args, device: Optional[Union[str, Tuple[str, str]]] = None, **kwargs
    ):
        super().__init__(*args, device=device, **kwargs)


class SoundEventWithResource(SoundEvent, ABC):
    """Base class for sound events with resource names attached"""

    def __init__(self, *args, resource: Optional[str] = None, **kwargs):
        super().__init__(*args, resource=resource, **kwargs)


class SoundPlaybackStartedEvent(SoundEventWithResource):
    """
    Event triggered when a new sound playback starts
    """


class SoundPlaybackStoppedEvent(SoundEventWithResource):
    """
    Event triggered when the sound playback stops
    """


class SoundPlaybackPausedEvent(SoundEventWithResource):
    """
    Event triggered when the sound playback pauses
    """


class SoundPlaybackResumedEvent(SoundEventWithResource):
    """
    Event triggered when the sound playback resumsed from a paused state
    """


class SoundRecordingStartedEvent(SoundEventWithResource):
    """
    Event triggered when a new recording starts
    """


class SoundRecordingStoppedEvent(SoundEventWithResource):
    """
    Event triggered when a sound recording stops
    """


class SoundRecordingPausedEvent(SoundEventWithResource):
    """
    Event triggered when a sound recording pauses
    """


class SoundRecordingResumedEvent(SoundEvent):
    """
    Event triggered when a sound recording resumes from a paused state
    """


# vim:sw=4:ts=4:et:



from multiprocessing import Process
from threading import RLock, Timer
from time import time

from platypush.context import get_bus
from platypush.message.event.flic import FlicButtonEvent
from platypush.plugins import RunnablePlugin

from .fliclib import FlicClient, ButtonConnectionChannel, ClickType


class FlicPlugin(RunnablePlugin):
    """
    This integration listens for events from the `Flic <https://flic.io/>`_
    smart buttons over Bluetooth.

    Requires:

        * **fliclib** (https://github.com/50ButtonsEach/fliclib-linux-hci).

    Clone the repository and follow the instructions in the README. This plugin
    requires:

        * The ``flicd`` daemon to be running - either on the same machine or on
          a machine that can be reached through the network.

        * The buttons to be paired with the device running the ``flicd`` daemon.
          The repository provides several scanners, but the easiest way is
          probably through the ``new_scan_wizard.py`` script.

    .. code-block:: bash

      # Clone the repository
      $ git clone https://github.com/50ButtonsEach/fliclib-linux-hci.git
      # Run the flid daemon
      $ [sudo] fliclib-linux-hci/bin/$(uname -m)/flicd -f /path/to/flicd.db &
      # Run the new_scan_wizard.py script to pair the buttons
      $ cd fliclib-linux-hci/clientlib/python
      $ python3 new_scan_wizard.py

    """

    _long_press_timeout = 0.3
    _btn_timeout = 0.5
    ShortPressEvent = "ShortPressEvent"
    LongPressEvent = "LongPressEvent"

    def __init__(
        self,
        server: str = 'localhost',
        long_press_timeout: float = _long_press_timeout,
        btn_timeout: float = _btn_timeout,
        **kwargs,
    ):
        """
        :param server: flicd server host (default: localhost)
        :param long_press_timeout: How long you should press a button for a
            press action to be considered "long press" (default: 0.3 secohds)
        :param btn_timeout: How long since the last button release before
            considering the user interaction completed (default: 0.5 seconds)
        """

        super().__init__(**kwargs)

        self.server = server
        self._client = None
        self._client_lock = RLock()

        self._long_press_timeout = long_press_timeout
        self._btn_timeout = btn_timeout
        self._btn_timer = None
        self._btn_addr = None
        self._down_pressed_time = None
        self._cur_sequence = []

    @property
    def client(self):
        with self._client_lock:
            if not self._client:
                self._client = FlicClient(self.server)
                self._client.get_info(self._received_info)
                self._client.on_new_verified_button = self._got_button

            return self._client

    def _got_button(self, bd_addr):
        cc = ButtonConnectionChannel(bd_addr)
        cc.on_button_up_or_down = (
            lambda channel, click_type, was_queued, time_diff: self._on_event(
                bd_addr, channel, click_type, was_queued, time_diff
            )
        )
        self.client.add_connection_channel(cc)

    def _received_info(self, items):
        for bd_addr in items["bd_addr_of_verified_buttons"]:
            self._got_button(bd_addr)

    def _on_btn_timeout(self):
        if self._btn_addr:
            get_bus().post(
                FlicButtonEvent(
                    btn_addr=self._btn_addr,
                    sequence=self._cur_sequence,
                )
            )

        self._cur_sequence = []

    # _ = channel
    # __ = time_diff
    def _on_event(self, bd_addr, _, click_type, was_queued, __):
        if was_queued and self._btn_addr:
            return

        if self._btn_timer:
            self._btn_timer.cancel()

        if click_type == ClickType.ButtonDown:
            self._down_pressed_time = time()
            return

        btn_event = self.ShortPressEvent
        if self._down_pressed_time:
            if time() - self._down_pressed_time >= self._long_press_timeout:
                btn_event = self.LongPressEvent
            self._down_pressed_time = None

        self._cur_sequence.append(btn_event)
        self._btn_addr = bd_addr
        self._btn_timer = Timer(self._btn_timeout, self._on_btn_timeout)
        self._btn_timer.start()

    def _processor(self):
        try:
            self.client.handle_events()
        except KeyboardInterrupt:
            pass

    def main(self):
        while not self.should_stop():
            proc = Process(target=self._processor, name='FlicProcessor')
            proc.start()
            self.wait_stop()

            if proc.is_alive():
                proc.terminate()
                proc.join(2)

            if proc.is_alive():
                self.logger.warning('Flic processor still alive after termination')
                proc.kill()


# vim:sw=4:ts=4:et:



from sqlalchemy import Column, Integer, ForeignKey, Float, String

from platypush.common.db import is_defined

from .devices import Device


if not is_defined('dimmer'):

    class Dimmer(Device):
        """
        This class models dimmer entities. A dimmer is any actionable entity
        with numeric values and an optional min/max range.
        """

        __tablename__ = 'dimmer'

        id = Column(
            Integer, ForeignKey(Device.id, ondelete='CASCADE'), primary_key=True
        )
        min = Column(Float)
        max = Column(Float)
        step = Column(Float, default=1.0)
        value = Column(Float)
        unit = Column(String)

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }



from dataclasses import dataclass
from typing import Any, Collection, Dict, Optional

from ._model import AccountConfig
from ._plugin import BaseMailPlugin, MailInPlugin, MailOutPlugin


@dataclass
class Account:
    """
    Models a mail account.
    """

    name: str
    poll_interval: float
    display_name: Optional[str] = None
    incoming: Optional[MailInPlugin] = None
    outgoing: Optional[MailOutPlugin] = None
    monitor_folders: Optional[Collection[str]] = None
    default: bool = False
    last_check: Optional[float] = None

    @classmethod
    def build(
        cls,
        name: str,
        timeout: float,
        poll_interval: float,
        display_name: Optional[str] = None,
        incoming: Optional[Dict[str, Any]] = None,
        outgoing: Optional[Dict[str, Any]] = None,
        monitor_folders: Optional[Collection[str]] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        password_cmd: Optional[str] = None,
        keyfile: Optional[str] = None,
        certfile: Optional[str] = None,
        access_token: Optional[str] = None,
        oauth_mechanism: Optional[str] = None,
        oauth_vendor: Optional[str] = None,
        default: bool = False,
        last_check: Optional[float] = None,
    ) -> 'Account':
        account_args = {
            'username': username,
            'password': password,
            'password_cmd': password_cmd,
            'access_token': access_token,
            'oauth_mechanism': oauth_mechanism,
            'oauth_vendor': oauth_vendor,
            'display_name': display_name,
        }

        in_plugin = None
        if incoming:
            server = incoming.pop('server', None)
            assert server, 'No server provided for incoming mail for account "{name}"'

            keyfile = incoming.pop('keyfile', keyfile)
            certfile = incoming.pop('certfile', certfile)
            account = AccountConfig(**{**account_args, **incoming})
            in_plugin = BaseMailPlugin.build(
                server=server,
                account=account,
                timeout=timeout,
                keyfile=keyfile,
                certfile=certfile,
            )

            assert isinstance(
                in_plugin, MailInPlugin
            ), 'Incoming mail plugin expected for account "{name}"'

        out_plugin = None
        if outgoing:
            server = outgoing.pop('server', None)
            assert server, 'No server provided for outgoing mail for account "{name}"'

            keyfile = outgoing.pop('keyfile', keyfile)
            certfile = outgoing.pop('certfile', certfile)
            account = AccountConfig(**{**account_args, **outgoing})
            out_plugin = BaseMailPlugin.build(
                server=server,
                account=account,
                timeout=timeout,
                keyfile=keyfile,
                certfile=certfile,
            )

            assert isinstance(
                out_plugin, MailOutPlugin
            ), 'Outgoing mail plugin expected for account "{name}"'

        return cls(
            name=name,
            display_name=display_name,
            incoming=in_plugin,
            outgoing=out_plugin,
            monitor_folders=monitor_folders,
            poll_interval=poll_interval,
            default=default,
            last_check=last_check,
        )


# vim:sw=4:ts=4:et:



import subprocess

from dataclasses import dataclass
from typing import Optional


@dataclass
class AccountConfig:
    """
    Model for a mail account configuration.
    """

    username: str
    password: Optional[str] = None
    password_cmd: Optional[str] = None
    access_token: Optional[str] = None
    oauth_mechanism: Optional[str] = None
    oauth_vendor: Optional[str] = None
    display_name: Optional[str] = None

    def __post_init__(self):
        """
        Ensure that at least one of password, password_cmd or access_token is provided.
        """
        assert (
            self.password or self.password_cmd or self.access_token
        ), 'No password, password_cmd or access_token provided'

    def get_password(self) -> str:
        """
        Get the password either from a provided string or from a password command.
        """
        if self.password_cmd:
            with subprocess.Popen(
                ['sh', '-c', self.password_cmd], stdout=subprocess.PIPE
            ) as proc:
                return proc.communicate()[0].decode()

        assert self.password, 'No password provided'
        return self.password


# vim:sw=4:ts=4:et:
